"""Trigger remote services on a vehicle."""

import asyncio
import datetime
import json
import logging
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

from bimmer_connected.api.client import MyBMWClient
from bimmer_connected.const import (
    REMOTE_SERVICE_POSITION_URL,
    REMOTE_SERVICE_STATUS_URL,
    REMOTE_SERVICE_URL,
    VEHICLE_CHARGING_PROFILE_SET_URL,
    VEHICLE_CHARGING_SETTINGS_SET_URL,
    VEHICLE_CHARGING_START_STOP_URL,
    VEHICLE_POI_URL,
)
from bimmer_connected.models import ChargingSettings, MyBMWRemoteServiceError, PointOfInterest, StrEnum
from bimmer_connected.utils import MyBMWJSONEncoder
from bimmer_connected.vehicle.charging_profile import (
    MAP_CHARGING_MODE_TO_REMOTE_SERVICE,
    ChargingMode,
    ChargingPreferences,
)
from bimmer_connected.vehicle.fuel_and_battery import ChargingState

if TYPE_CHECKING:
    from bimmer_connected.vehicle import MyBMWVehicle

TIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"

_LOGGER = logging.getLogger(__name__)

#: time in seconds between polling updates on the status of a remote service
_POLLING_CYCLE = 3

#: maximum number of seconds to wait for the server to return a positive answer
_POLLING_TIMEOUT = 240


class ExecutionState(StrEnum):
    """Enumeration of possible states of the execution of a remote service."""

    INITIATED = "INITIATED"
    PENDING = "PENDING"
    DELIVERED = "DELIVERED"
    EXECUTED = "EXECUTED"
    ERROR = "ERROR"
    IGNORED = "IGNORED"
    UNKNOWN = "UNKNOWN"


class Services(StrEnum):
    """Enumeration of possible services to be executed."""

    LIGHT_FLASH = "light-flash"
    VEHICLE_FINDER = "vehicle-finder"
    DOOR_LOCK = "door-lock"
    DOOR_UNLOCK = "door-unlock"
    HORN = "horn-blow"
    AIR_CONDITIONING = "climate-now"
    CHARGE_START = "start-charging"
    CHARGE_STOP = "stop-charging"
    CHARGING_SETTINGS = "CHARGING_SETTINGS"
    CHARGING_PROFILE = "CHARGING_PROFILE"
    SEND_POI = "SEND_POI"


# Non-default remote services URLs
SERVICE_URLS = {
    Services.CHARGING_SETTINGS: VEHICLE_CHARGING_SETTINGS_SET_URL,
    Services.CHARGING_PROFILE: VEHICLE_CHARGING_PROFILE_SET_URL,
    Services.SEND_POI: VEHICLE_POI_URL,
    Services.CHARGE_START: VEHICLE_CHARGING_START_STOP_URL,
    Services.CHARGE_STOP: VEHICLE_CHARGING_START_STOP_URL,
}

CHARGING_MODE_TO_CHARGING_PREFERENCE = {
    ChargingMode.IMMEDIATE_CHARGING: ChargingPreferences.NO_PRESELECTION,
    ChargingMode.DELAYED_CHARGING: ChargingPreferences.CHARGING_WINDOW,
}


class RemoteServiceStatus:
    """Wraps the status of the execution of a remote service."""

    def __init__(self, response: dict, event_id: Optional[str] = None):
        """Construct a new object from a dict."""
        status = None
        if "eventStatus" in response:
            status = response.get("eventStatus")

        self.state = ExecutionState(status or "UNKNOWN")
        self.details = response
        self.event_id = event_id


class RemoteServices:
    """Trigger remote services on a vehicle."""

    def __init__(self, vehicle: "MyBMWVehicle"):
        self._account = vehicle.account
        self._vehicle = vehicle

    async def trigger_remote_service(
        self, service_id: Services, params: Optional[Dict] = None, data: Any = None, refresh: bool = False
    ) -> RemoteServiceStatus:
        """Trigger a remote service and wait for the result."""

        # Check if service requires a specific url and add all required parameters
        url = SERVICE_URLS.get(service_id, REMOTE_SERVICE_URL)
        url = url.format(vin=self._vehicle.vin, service_type=service_id.value)

        # Trigger service and get event id
        async with MyBMWClient(self._account.config, brand=self._vehicle.brand) as client:
            response = await client.post(
                url,
                headers={"content-type": "application/json"} if data else None,
                params=params,
                content=json.dumps(data, cls=MyBMWJSONEncoder) if data else None,
            )
            event_id = response.json().get("eventId") if response.content else None

            # Get status via event_id or assume successful execution as HTTP errors would raise exceptions before
            status = (
                await self._block_until_done(client, event_id)
                if event_id
                else RemoteServiceStatus({"eventStatus": "EXECUTED"})
            )

        # If vehicle data needs to be refresh, wait 2 times polling cycle and refresh completely
        if refresh:
            await asyncio.sleep(_POLLING_CYCLE * 2)
            await self._account.get_vehicles()

        return status

    async def _get_remote_service_status(self, client: MyBMWClient, event_id: str) -> RemoteServiceStatus:
        """Return execution status of the last remote service that was triggered."""

        _LOGGER.debug("getting remote service status for '%s'", event_id)
        url = REMOTE_SERVICE_STATUS_URL.format(vin=self._vehicle.vin, event_id=event_id)
        async with MyBMWClient(self._account.config, brand=self._vehicle.brand) as client:
            response = await client.post(url)
        return RemoteServiceStatus(response.json(), event_id=event_id)

    async def _block_until_done(self, client: MyBMWClient, event_id: str) -> RemoteServiceStatus:
        """Keep polling the server until we get a final answer.

        :raises TimeoutError: if there is no final answer before _POLLING_TIMEOUT
        """

        fail_after = datetime.datetime.now() + datetime.timedelta(seconds=_POLLING_TIMEOUT)
        while datetime.datetime.now() < fail_after:
            await asyncio.sleep(_POLLING_CYCLE)
            status = await self._get_remote_service_status(client, event_id)
            _LOGGER.debug("current state of '%s' is: %s", event_id, status.state.value)
            if status.state == ExecutionState.ERROR:
                raise MyBMWRemoteServiceError(
                    f"Remote service failed with state '{status.state}'. Response: {status.details}"
                )
            if status.state not in [ExecutionState.UNKNOWN, ExecutionState.PENDING, ExecutionState.DELIVERED]:
                return status
        raise MyBMWRemoteServiceError(
            f"Did not receive remote service result for '{event_id}' in {_POLLING_TIMEOUT} seconds. "
            f"Current state: {status.state.value}"
        )

    async def trigger_remote_light_flash(self) -> RemoteServiceStatus:
        """Trigger the vehicle to flash its headlights."""
        if not self._vehicle.is_remote_lights_enabled:
            raise ValueError(f"Vehicle does not support remote service '{Services.LIGHT_FLASH.value}'.")
        return await self.trigger_remote_service(Services.LIGHT_FLASH)

    async def trigger_remote_door_lock(self) -> RemoteServiceStatus:
        """Trigger the vehicle to lock its doors."""
        if not self._vehicle.is_remote_lock_enabled:
            raise ValueError(f"Vehicle does not support remote service '{Services.DOOR_LOCK.value}'.")
        return await self.trigger_remote_service(Services.DOOR_LOCK, refresh=True)

    async def trigger_remote_door_unlock(self) -> RemoteServiceStatus:
        """Trigger the vehicle to unlock its doors."""
        if not self._vehicle.is_remote_unlock_enabled:
            raise ValueError(f"Vehicle does not support remote service '{Services.DOOR_UNLOCK.value}'.")
        return await self.trigger_remote_service(Services.DOOR_UNLOCK, refresh=True)

    async def trigger_remote_horn(self) -> RemoteServiceStatus:
        """Trigger the vehicle to sound its horn."""
        if not self._vehicle.is_remote_horn_enabled:
            raise ValueError(f"Vehicle does not support remote service '{Services.HORN.value}'.")
        return await self.trigger_remote_service(Services.HORN)

    async def trigger_charge_start(self) -> RemoteServiceStatus:
        """Trigger the vehicle to start charging."""
        if not self._vehicle.is_remote_charge_start_enabled:
            raise ValueError(f"Vehicle does not support remote service '{Services.CHARGE_START.value}'.")

        if not self._vehicle.fuel_and_battery.is_charger_connected:
            _LOGGER.warning("Charger not connected, cannot start charging.")
            return RemoteServiceStatus({"eventStatus": "IGNORED"})

        return await self.trigger_remote_service(Services.CHARGE_START, refresh=True)

    async def trigger_charge_stop(self) -> RemoteServiceStatus:
        """Trigger the vehicle to stop charging."""
        if not self._vehicle.is_remote_charge_stop_enabled:
            raise ValueError(f"Vehicle does not support remote service '{Services.CHARGE_STOP.value}'.")

        if not self._vehicle.fuel_and_battery.is_charger_connected:
            _LOGGER.warning("Charger not connected, cannot stop charging.")
            return RemoteServiceStatus({"eventStatus": "IGNORED"})
        if self._vehicle.fuel_and_battery.charging_status != ChargingState.CHARGING:
            _LOGGER.warning("Vehicle not charging, cannot stop charging.")
            return RemoteServiceStatus({"eventStatus": "IGNORED"})

        return await self.trigger_remote_service(Services.CHARGE_STOP, refresh=True)

    async def trigger_remote_air_conditioning(self) -> RemoteServiceStatus:
        """Trigger the air conditioning to start."""
        if not self._vehicle.is_remote_climate_start_enabled:
            raise ValueError(
                f"Vehicle does not support remote service '{Services.AIR_CONDITIONING.value}' action 'START'."
            )
        return await self.trigger_remote_service(Services.AIR_CONDITIONING, params={"action": "START"}, refresh=True)

    async def trigger_remote_air_conditioning_stop(self) -> RemoteServiceStatus:
        """Trigger the air conditioning to stop."""
        if not self._vehicle.is_remote_climate_stop_enabled:
            raise ValueError(
                f"Vehicle does not support remote service '{Services.AIR_CONDITIONING.value}' action 'STOP'."
            )
        return await self.trigger_remote_service(Services.AIR_CONDITIONING, params={"action": "STOP"}, refresh=True)

    async def trigger_charging_settings_update(
        self, target_soc: Optional[int] = None, ac_limit: Optional[int] = None
    ) -> RemoteServiceStatus:
        """Update the charging settings on the vehicle."""

        if target_soc and not self._vehicle.is_remote_set_target_soc_enabled:
            raise ValueError("Vehicle does not support setting target SoC.")
        if target_soc and (
            not isinstance(target_soc, int) or target_soc < 20 or target_soc > 100 or target_soc % 5 != 0
        ):
            raise ValueError("Target SoC must be an integer between 20 and 100 that is a multiple of 5.")
        if ac_limit:
            if (
                not self._vehicle.is_remote_set_ac_limit_enabled
                or not self._vehicle.charging_profile
                or not self._vehicle.charging_profile.ac_available_limits
            ):
                raise ValueError("Vehicle does not support setting AC Limit.")
            if not isinstance(ac_limit, int) or ac_limit not in self._vehicle.charging_profile.ac_available_limits:
                raise ValueError("AC Limit must be an integer and in `charging_profile.ac_available_limits`.")

        return await self.trigger_remote_service(
            Services.CHARGING_SETTINGS,
            data=ChargingSettings(chargingTarget=target_soc, acLimitValue=ac_limit),
            refresh=True,
        )

    async def trigger_charging_profile_update(
        self, charging_mode: Optional[ChargingMode] = None, precondition_climate: Optional[bool] = None
    ) -> RemoteServiceStatus:
        """Update the charging profile on the vehicle."""

        if not self._vehicle.is_charging_plan_supported or not self._vehicle.charging_profile:
            raise ValueError("Vehicle does not support setting charging profile.")

        target_charging_profile = self._vehicle.charging_profile.format_for_remote_service()

        if charging_mode and not charging_mode == ChargingMode.UNKNOWN:
            target_charging_profile["chargingMode"]["type"] = MAP_CHARGING_MODE_TO_REMOTE_SERVICE[charging_mode]
            target_charging_profile["chargingMode"]["chargingPreference"] = CHARGING_MODE_TO_CHARGING_PREFERENCE[
                charging_mode
            ].value

        if precondition_climate is not None:
            target_charging_profile["isPreconditionForDepartureActive"] = precondition_climate

        return await self.trigger_remote_service(
            Services.CHARGING_PROFILE,
            data=target_charging_profile,
            refresh=True,
        )

    async def trigger_send_poi(self, poi: Union[PointOfInterest, Dict]) -> RemoteServiceStatus:
        """Send a PointOfInterest to the vehicle.

        :param poi: A PointOfInterest containing at least 'lat' and 'lon' and optionally
            'name', 'street', 'city', 'postalCode', 'country'
        """
        if not self._vehicle.is_remote_sendpoi_enabled:
            raise ValueError(f"Vehicle does not support remote service '{Services.SEND_POI.value}'.")

        if isinstance(poi, Dict):
            poi = PointOfInterest(**poi)

        return await self.trigger_remote_service(
            Services.SEND_POI,
            data={
                "location": poi,
                "vin": self._vehicle.vin,
            },
        )

    async def trigger_remote_vehicle_finder(self) -> RemoteServiceStatus:
        """Trigger the vehicle finder."""
        if not self._vehicle.is_vehicle_tracking_enabled:
            raise ValueError(f"Vehicle does not support remote service '{Services.VEHICLE_FINDER.value}'.")

        status = await self.trigger_remote_service(Services.VEHICLE_FINDER)
        result = await self._get_event_position(status.event_id)
        self._vehicle.vehicle_location.set_remote_service_position(result)
        return status

    async def _get_event_position(self, event_id) -> Dict:
        url = REMOTE_SERVICE_POSITION_URL.format(event_id=event_id)
        if not self._account.config.observer_position:
            return {
                "errorDetails": {
                    "title": "Unknown position",
                    "description": "Set observer position to retrieve vehicle coordinates!",
                }
            }
        async with MyBMWClient(self._account.config, brand=self._vehicle.brand) as client:
            response = await client.post(
                url,
                headers={
                    "latitude": str(self._account.config.observer_position.latitude),
                    "longitude": str(self._account.config.observer_position.longitude),
                },
            )
        return response.json()



"""Tests for API that are not covered by other tests."""
import json

import httpx
import pytest
import respx

from bimmer_connected.account import MyBMWAccount
from bimmer_connected.api.authentication import get_retry_wait_time
from bimmer_connected.api.regions import get_region_from_name, valid_regions
from bimmer_connected.api.utils import anonymize_data
from bimmer_connected.utils import log_response_store_to_file

from . import (
    RESPONSE_DIR,
    TEST_PASSWORD,
    TEST_REGION,
    TEST_USERNAME,
    VIN_G26,
    get_fingerprint_charging_settings_count,
    get_fingerprint_state_count,
    # get_fingerprint_count,
    load_response,
)


def test_valid_regions():
    """Test valid regions."""
    assert ["north_america", "china", "rest_of_world"] == valid_regions()


def test_unknown_region():
    """Test unknown region."""
    with pytest.raises(ValueError):
        get_region_from_name("unknown")


def test_anonymize_data():
    """Test anonymization function."""
    test_dict = {
        "vin": "WBA000000SECRET01",
        "a sub-dict": {
            "lat": 666,
            "lon": 666,
            "heading": 666,
        },
        "licensePlate": "secret",
        "public": "public_data",
        "a_list": [
            {"vin": "WBA000000SECRET01"},
            {
                "lon": 666,
                "public": "more_public_data",
            },
        ],
        "b_list": ["a", "b"],
        "empty_list": [],
    }
    anon_text = json.dumps(anonymize_data(test_dict))
    assert "SECRET" not in anon_text
    assert "secret" not in anon_text
    assert "666" not in anon_text
    assert "public_data" in anon_text
    assert "more_public_data" in anon_text


@pytest.mark.asyncio
async def test_storing_fingerprints(tmp_path, bmw_fixture: respx.Router, bmw_log_all_responses):
    """Test storing fingerprints to file."""

    # We need to remove the existing state route first and add it back later as otherwise our error call is never
    # matched (respx matches by order of routes and we don't replace the existing one)
    state_route = bmw_fixture.routes.pop("state")
    bmw_fixture.get("/eadrax-vcs/v4/vehicles/state", headers={"bmw-vin": VIN_G26}).respond(
        500, text=load_response(RESPONSE_DIR / "auth" / "auth_error_internal_error.txt")
    )
    bmw_fixture.routes.add(state_route, "state")

    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION, log_responses=True)
    await account.get_vehicles()

    log_response_store_to_file(account.get_stored_responses(), tmp_path)

    files = list(tmp_path.iterdir())
    json_files = [f for f in files if f.suffix == ".json"]
    txt_files = [f for f in files if f.suffix == ".txt"]

    assert len(json_files) == (get_fingerprint_state_count() + get_fingerprint_charging_settings_count() + 2 - 2)
    assert len(txt_files) == 1


@pytest.mark.asyncio
async def test_fingerprint_deque(monkeypatch: pytest.MonkeyPatch, bmw_fixture: respx.Router):
    """Test storing fingerprints to file."""
    # Increase length of response store for local testing

    # temp_store = deque(maxlen=100)
    # monkeypatch.setattr("bimmer_connected.api.client.RESPONSE_STORE", temp_store)
    # monkeypatch.setattr("bimmer_connected.account.RESPONSE_STORE", temp_store)

    # Prepare Number of good responses (vehicle states, charging settings per vehicle)
    # # and 2x vehicle list
    # json_count = get_fingerprint_state_count() + get_fingerprint_charging_settings_count() + 2

    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION, log_responses=True)
    await account.get_vehicles()
    await account.get_vehicles()

    # More than 10 calls were made, but only last 10 are stored
    assert len([c for c in bmw_fixture.calls if c.request.url.path.startswith("/eadrax-vcs")]) > 10
    assert len(account.get_stored_responses()) == 10

    # Stored responses are reset
    account.config.set_log_responses(False)
    assert len(account.get_stored_responses()) == 0

    # No new responses are getting added
    await account.get_vehicles()
    assert len(account.get_stored_responses()) == 0

    # Get responses again
    account.config.set_log_responses(True)
    await account.get_vehicles()
    assert len(account.get_stored_responses()) == 10


def test_get_retry_wait_time():
    """Test extraction of retry wait time."""

    # Parsing correctly
    r = httpx.Response(429, json={"statusCode": 429, "message": "Rate limit is exceeded. Try again in 1 seconds."})
    assert get_retry_wait_time(r) == 2

    # No number found
    r = httpx.Response(429, json={"statusCode": 429, "message": "Rate limit is exceeded."})
    assert get_retry_wait_time(r) == 4

    # No JSON response
    r = httpx.Response(429, text="Rate limit is exceeded.")
    assert get_retry_wait_time(r) == 4



"""Tests for MyBMWAccount."""

import datetime
import logging
from pathlib import Path
from unittest import mock

import httpx
import pytest
import respx

from bimmer_connected.account import MyBMWAccount
from bimmer_connected.api.authentication import MyBMWAuthentication, MyBMWLoginRetry
from bimmer_connected.api.client import MyBMWClient
from bimmer_connected.api.regions import get_region_from_name
from bimmer_connected.const import ATTR_CAPABILITIES
from bimmer_connected.models import GPSPosition, MyBMWAPIError, MyBMWAuthError, MyBMWQuotaError

from . import (
    RESPONSE_DIR,
    TEST_PASSWORD,
    TEST_REGION,
    TEST_REGION_STRING,
    TEST_USERNAME,
    VIN_G26,
    VIN_I20,
    get_fingerprint_charging_settings_count,
    get_fingerprint_state_count,
    load_response,
)
from .conftest import prepare_account_with_vehicles


@pytest.mark.asyncio
async def test_login_row_na(bmw_fixture: respx.Router):
    """Test the login flow."""
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, get_region_from_name(TEST_REGION_STRING))
    await account.get_vehicles()
    assert account is not None


@pytest.mark.asyncio
async def test_login_refresh_token_row_na_expired(bmw_fixture: respx.Router):
    """Test the login flow using refresh_token."""
    with mock.patch("bimmer_connected.api.authentication.EXPIRES_AT_OFFSET", datetime.timedelta(seconds=30000)):
        account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, get_region_from_name(TEST_REGION_STRING))
        await account.get_vehicles()

        with mock.patch(
            "bimmer_connected.api.authentication.MyBMWAuthentication._refresh_token_row_na",
            wraps=account.config.authentication._refresh_token_row_na,
        ) as mock_listener:
            mock_listener.reset_mock()
            await account.get_vehicles()

            # Should not be called at all, as expiry date is not checked anymore
            assert mock_listener.call_count == 0
            assert account.config.authentication.refresh_token is not None


@pytest.mark.asyncio
async def test_login_refresh_token_row_na_401(bmw_fixture: respx.Router):
    """Test the login flow using refresh_token."""

    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, get_region_from_name(TEST_REGION_STRING))
    await account.get_vehicles()

    with mock.patch(
        "bimmer_connected.api.authentication.MyBMWAuthentication._refresh_token_row_na",
        wraps=account.config.authentication._refresh_token_row_na,
    ) as mock_listener:
        bmw_fixture.get("/eadrax-vcs/v4/vehicles/state").mock(
            side_effect=[httpx.Response(401), *([httpx.Response(200, json={ATTR_CAPABILITIES: {}})] * 10)]
        )
        mock_listener.reset_mock()
        await account.get_vehicles()

        assert mock_listener.call_count == 1
        assert account.config.authentication.refresh_token is not None


@pytest.mark.asyncio
async def test_login_refresh_token_row_na_invalid(caplog, bmw_fixture: respx.Router):
    """Test the login flow using refresh_token."""
    bmw_fixture.post("/gcdm/oauth/token").mock(
        side_effect=[
            httpx.Response(400),
            httpx.Response(200, json=load_response(RESPONSE_DIR / "auth" / "auth_token.json")),
        ]
    )

    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, get_region_from_name(TEST_REGION_STRING))
    account.set_refresh_token("INVALID")

    caplog.set_level(logging.DEBUG)
    await account.get_vehicles()

    debug_messages = [r.message for r in caplog.records if r.name.startswith("bimmer_connected")]
    assert "Authenticating with refresh token for North America & Rest of World." in debug_messages
    assert "Unable to get access token using refresh token, falling back to username/password." in debug_messages
    assert "Authenticating with MyBMW flow for North America & Rest of World." in debug_messages


@pytest.mark.asyncio
async def test_login_china(bmw_fixture: respx.Router):
    """Test the login flow for region `china`."""
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, get_region_from_name("china"))
    await account.get_vehicles()
    assert account is not None


@pytest.mark.asyncio
async def test_login_refresh_token_china_expired(bmw_fixture: respx.Router):
    """Test the login flow using refresh_token  for region `china`."""
    with mock.patch("bimmer_connected.api.authentication.EXPIRES_AT_OFFSET", datetime.timedelta(seconds=30000)):
        account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, get_region_from_name("china"))
        await account.get_vehicles()

        with mock.patch(
            "bimmer_connected.api.authentication.MyBMWAuthentication._refresh_token_china",
            wraps=account.config.authentication._refresh_token_china,
        ) as mock_listener:
            mock_listener.reset_mock()
            await account.get_vehicles()

            # Should not be called at all, as expiry date is not checked anymore
            assert mock_listener.call_count == 0
            assert account.config.authentication.refresh_token is not None


@pytest.mark.asyncio
async def test_login_refresh_token_china_401(bmw_fixture: respx.Router):
    """Test the login flow using refresh_token."""
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, get_region_from_name("china"))
    await account.get_vehicles()

    with mock.patch(
        "bimmer_connected.api.authentication.MyBMWAuthentication._refresh_token_china",
        wraps=account.config.authentication._refresh_token_china,
    ) as mock_listener:
        bmw_fixture.get("/eadrax-vcs/v4/vehicles/state").mock(
            side_effect=[httpx.Response(401), *([httpx.Response(200, json={ATTR_CAPABILITIES: {}})] * 10)]
        )
        mock_listener.reset_mock()
        await account.get_vehicles()

        assert mock_listener.call_count == 1
        assert account.config.authentication.refresh_token is not None


@pytest.mark.asyncio
async def test_login_refresh_token_china_invalid(caplog, bmw_fixture: respx.Router):
    """Test the login flow using refresh_token."""
    bmw_fixture.post("/eadrax-coas/v2/oauth/token").mock(
        side_effect=[
            httpx.Response(400),
            httpx.Response(200, json=load_response(RESPONSE_DIR / "auth" / "auth_token.json")),
        ]
    )

    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, get_region_from_name("china"))
    account.set_refresh_token("INVALID")

    caplog.set_level(logging.DEBUG)
    await account.get_vehicles()

    debug_messages = [r.message for r in caplog.records if r.name.startswith("bimmer_connected")]
    assert "Authenticating with refresh token for China." in debug_messages
    assert "Unable to get access token using refresh token, falling back to username/password." in debug_messages
    assert "Authenticating with MyBMW flow for China." in debug_messages


@pytest.mark.asyncio
async def test_vehicles(bmw_fixture: respx.Router):
    """Test the login flow."""
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, get_region_from_name("china"))
    await account.get_vehicles()

    assert account.config.authentication.access_token is not None
    assert get_fingerprint_state_count() == len(account.vehicles)

    vehicle = account.get_vehicle(VIN_G26)
    assert vehicle is not None
    assert VIN_G26 == vehicle.vin

    assert account.get_vehicle("invalid_vin") is None


@pytest.mark.asyncio
async def test_vehicle_init(bmw_fixture: respx.Router):
    """Test vehicle initialization."""
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)
    with mock.patch(
        "bimmer_connected.account.MyBMWAccount._init_vehicles",
        wraps=account._init_vehicles,
    ) as mock_listener:
        mock_listener.reset_mock()

        # First call on init
        await account.get_vehicles()
        assert len(account.vehicles) == get_fingerprint_state_count()

        # No call to _init_vehicles()
        await account.get_vehicles()
        assert len(account.vehicles) == get_fingerprint_state_count()

        # Second, forced call _init_vehicles()
        await account.get_vehicles(force_init=True)
        assert len(account.vehicles) == get_fingerprint_state_count()

        assert mock_listener.call_count == 2


@pytest.mark.asyncio
async def test_invalid_password(bmw_fixture: respx.Router):
    """Test parsing the results of an invalid password."""
    bmw_fixture.post("/gcdm/oauth/authenticate").respond(
        401, json=load_response(RESPONSE_DIR / "auth" / "auth_error_wrong_password.json")
    )
    with pytest.raises(MyBMWAuthError):
        account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)
        await account.get_vehicles()


@pytest.mark.asyncio
async def test_invalid_password_china(bmw_fixture: respx.Router):
    """Test parsing the results of an invalid password."""
    bmw_fixture.post("/eadrax-coas/v2/login/pwd").respond(
        422, json=load_response(RESPONSE_DIR / "auth" / "auth_cn_login_error.json")
    )
    with pytest.raises(MyBMWAPIError):
        account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, get_region_from_name("china"))
        await account.get_vehicles()


@pytest.mark.asyncio
async def test_server_error(bmw_fixture: respx.Router):
    """Test parsing the results of a server error."""
    bmw_fixture.post("/gcdm/oauth/authenticate").respond(
        500, text=load_response(RESPONSE_DIR / "auth" / "auth_error_internal_error.txt")
    )
    with pytest.raises(MyBMWAPIError):
        account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)
        await account.get_vehicles()


@pytest.mark.asyncio
async def test_vehicle_search_case(bmw_fixture: respx.Router):
    """Check if the search for the vehicle by VIN is NOT case sensitive."""
    account = await prepare_account_with_vehicles()

    vin = account.vehicles[1].vin
    assert vin == account.get_vehicle(vin).vin
    assert vin == account.get_vehicle(vin.lower()).vin
    assert vin == account.get_vehicle(vin.upper()).vin


@pytest.mark.asyncio
async def test_get_fingerprints(monkeypatch: pytest.MonkeyPatch, bmw_fixture: respx.Router, bmw_log_all_responses):
    """Test getting fingerprints."""

    # Prepare Number of good responses (vehicle states, charging settings per vehicle)
    # and 2x vehicle list
    json_count = get_fingerprint_state_count() + get_fingerprint_charging_settings_count() + 2

    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION, log_responses=True)
    await account.get_vehicles()

    # This should have been successful
    filenames = [Path(f.filename) for f in account.get_stored_responses()]
    json_files = [f for f in filenames if f.suffix == ".json"]
    txt_files = [f for f in filenames if f.suffix == ".txt"]

    assert len(json_files) == json_count  # all good
    assert len(txt_files) == 0  # no errors

    # Now we simulate an error for a single vehicle
    # We need to remove the existing state route first and add it back later as otherwise our error call is never
    # matched (respx matches by order of routes and we don't replace the existing one)
    state_route = bmw_fixture.routes.pop("state")
    bmw_fixture.get("/eadrax-vcs/v4/vehicles/state", headers={"bmw-vin": VIN_G26}).respond(
        500, text=load_response(RESPONSE_DIR / "auth" / "auth_error_internal_error.txt")
    )
    bmw_fixture.routes.add(state_route, "state")

    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION, log_responses=True)
    await account.get_vehicles()

    filenames = [Path(f.filename) for f in account.get_stored_responses()]
    json_files = [f for f in filenames if f.suffix == ".json"]
    txt_files = [f for f in filenames if f.suffix == ".txt"]

    assert len(json_files) == json_count - 2  # missing on 1 state and 1 charging setting
    assert len(txt_files) == 1  # error message from state, charging setting was not loaded anymore


@pytest.mark.asyncio
async def test_set_observer_value(bmw_fixture: respx.Router):
    """Test set_observer_position with valid arguments."""
    account = await prepare_account_with_vehicles()

    account.set_observer_position(1.0, 2.0)

    assert account.config.observer_position == GPSPosition(1.0, 2.0)


@pytest.mark.asyncio
async def test_set_observer_not_set(bmw_fixture: respx.Router):
    """Test set_observer_position with no arguments."""
    account = await prepare_account_with_vehicles()

    assert account.config.observer_position is None

    account.set_observer_position(17.99, 179.9)

    assert account.config.observer_position == GPSPosition(17.99, 179.9)


@pytest.mark.asyncio
async def test_set_observer_invalid_values(bmw_fixture: respx.Router):
    """Test set_observer_position with invalid arguments."""
    account = await prepare_account_with_vehicles()

    with pytest.raises(TypeError):
        account.set_observer_position(None, 2.0)

    with pytest.raises(TypeError):
        account.set_observer_position(1.0, None)

    with pytest.raises(TypeError):
        account.set_observer_position(1.0, "16.0")


@pytest.mark.asyncio
async def test_set_use_metric_units(caplog):
    """Test (deprecated) use_metrics_units flag."""

    # Default
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)
    assert len(caplog.records) == 0
    metric_client = MyBMWClient(account.config)
    assert metric_client.generate_default_header()["bmw-units-preferences"] == "d=KM;v=L"

    # Set to true
    caplog.clear()
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION, use_metric_units=True)
    assert len(caplog.records) == 1
    metric_client = MyBMWClient(account.config)
    assert metric_client.generate_default_header()["bmw-units-preferences"] == "d=KM;v=L"

    # Set to false
    caplog.clear()
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION, use_metric_units=True)
    assert len(caplog.records) == 1
    metric_client = MyBMWClient(account.config)
    assert metric_client.generate_default_header()["bmw-units-preferences"] == "d=KM;v=L"


@pytest.mark.asyncio
async def test_refresh_token_getset(bmw_fixture: respx.Router):
    """Test getting/setting the refresh_token and gcid."""
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)
    assert account.refresh_token is None
    await account.get_vehicles()
    assert account.refresh_token == "another_token_string"
    assert account.gcid is None

    account.set_refresh_token("new_refresh_token")
    assert account.refresh_token == "new_refresh_token"
    assert account.gcid is None

    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, get_region_from_name("china"))
    account.set_refresh_token("new_refresh_token", "dummy_gcid")
    assert account.refresh_token == "new_refresh_token"
    assert account.gcid == "dummy_gcid"
    await account.get_vehicles()
    assert account.refresh_token == "another_token_string"
    assert account.gcid == "DUMMY"


@pytest.mark.asyncio
async def test_429_retry_ok_login(caplog, bmw_fixture: respx.Router):
    """Test the login flow using refresh_token."""
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)

    json_429 = {"statusCode": 429, "message": "Rate limit is exceeded. Try again in 2 seconds."}

    bmw_fixture.get("/eadrax-ucs/v1/presentation/oauth/config").mock(
        side_effect=[
            httpx.Response(429, json=json_429),
            httpx.Response(429, json=json_429),
            httpx.Response(200, json=load_response(RESPONSE_DIR / "auth" / "oauth_config.json")),
        ]
    )
    caplog.set_level(logging.DEBUG)

    with mock.patch("asyncio.sleep", new_callable=mock.AsyncMock):
        await account.get_vehicles()

    log_429 = [
        r
        for r in caplog.records
        if r.module == "authentication" and "seconds due to 429 Too Many Requests" in r.message
    ]
    assert len(log_429) == 2


@pytest.mark.asyncio
async def test_429_retry_raise_login(caplog, bmw_fixture: respx.Router):
    """Test the login flow using refresh_token."""
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)

    json_429 = {"statusCode": 429, "message": "Rate limit is exceeded. Try again in 2 seconds."}

    bmw_fixture.get("/eadrax-ucs/v1/presentation/oauth/config").mock(return_value=httpx.Response(429, json=json_429))
    caplog.set_level(logging.DEBUG)

    with mock.patch("asyncio.sleep", new_callable=mock.AsyncMock):
        with pytest.raises(MyBMWAPIError):
            await account.get_vehicles()

    log_429 = [
        r
        for r in caplog.records
        if r.module == "authentication" and "seconds due to 429 Too Many Requests" in r.message
    ]
    assert len(log_429) == 3


@pytest.mark.asyncio
async def test_429_retry_ok_vehicles(caplog, bmw_fixture: respx.Router):
    """Test waiting on 429 for vehicles."""
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)

    json_429 = {"statusCode": 429, "message": "Rate limit is exceeded. Try again in 2 seconds."}

    bmw_fixture.get("/eadrax-vcs/v4/vehicles").mock(
        side_effect=[
            httpx.Response(429, json=json_429),
            httpx.Response(429, json=json_429),
            *[httpx.Response(200, json=[])] * 2,
        ]
    )
    caplog.set_level(logging.DEBUG)

    with mock.patch("asyncio.sleep", new_callable=mock.AsyncMock):
        await account.get_vehicles()

    log_429 = [
        r
        for r in caplog.records
        if r.module == "authentication" and "seconds due to 429 Too Many Requests" in r.message
    ]
    assert len(log_429) == 2


@pytest.mark.asyncio
async def test_429_retry_raise_vehicles(caplog, bmw_fixture: respx.Router):
    """Test waiting on 429 for vehicles and fail if it happens too often."""
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)

    json_429 = {"statusCode": 429, "message": "Rate limit is exceeded. Try again in 2 seconds."}

    bmw_fixture.get("/eadrax-vcs/v4/vehicles").mock(return_value=httpx.Response(429, json=json_429))
    caplog.set_level(logging.DEBUG)

    with mock.patch("asyncio.sleep", new_callable=mock.AsyncMock):
        with pytest.raises(MyBMWQuotaError):
            await account.get_vehicles()

    log_429 = [
        r
        for r in caplog.records
        if r.module == "authentication" and "seconds due to 429 Too Many Requests" in r.message
    ]
    assert len(log_429) == 3


@pytest.mark.asyncio
async def test_429_retry_with_login_ok_vehicles(bmw_fixture: respx.Router):
    """Test the login flow but experiencing a 429 first."""
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)

    json_429 = {"statusCode": 429, "message": "Rate limit is exceeded. Try again in 2 seconds."}

    bmw_fixture.get("/eadrax-vcs/v4/vehicles").mock(
        side_effect=[
            httpx.Response(429, json=json_429),
            httpx.Response(429, json=json_429),
            *[httpx.Response(200, json=[])] * 2,
        ]
    )

    with mock.patch("asyncio.sleep", new_callable=mock.AsyncMock):
        await account.get_vehicles()


@pytest.mark.asyncio
async def test_429_retry_with_login_raise_vehicles(bmw_fixture: respx.Router):
    """Test the error handling, experiencing a 429, 401 and another two 429."""
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)

    json_429 = {"statusCode": 429, "message": "Rate limit is exceeded. Try again in 2 seconds."}

    bmw_fixture.get("/eadrax-vcs/v4/vehicles").mock(
        side_effect=[
            httpx.Response(429, json=json_429),
            httpx.Response(401),
            httpx.Response(429, json=json_429),
            httpx.Response(429, json=json_429),
            httpx.Response(429, json=json_429),
            httpx.Response(429, json=json_429),
            httpx.Response(429, json=json_429),
        ]
    )

    with mock.patch("asyncio.sleep", new_callable=mock.AsyncMock):
        with pytest.raises(MyBMWQuotaError):
            await account.get_vehicles()


@pytest.mark.asyncio
async def test_multiple_401(bmw_fixture: respx.Router):
    """Test the error handling, when multiple 401 are received in sequence."""
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)

    bmw_fixture.get("/eadrax-vcs/v4/vehicles").mock(
        side_effect=[
            httpx.Response(401),
            httpx.Response(401),
        ]
    )

    with mock.patch("asyncio.sleep", new_callable=mock.AsyncMock):
        with pytest.raises(MyBMWAuthError):
            await account.get_vehicles()


@pytest.mark.asyncio
async def test_401_after_429_ok(bmw_fixture: respx.Router):
    """Test the error handling, when a 401 is received after exactly 3 429."""
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)
    await account.get_vehicles()

    json_429 = {"statusCode": 429, "message": "Rate limit is exceeded. Try again in 2 seconds."}

    # Recover after 3 429 and 1 401
    bmw_fixture.get("/eadrax-vcs/v4/vehicles").mock(
        side_effect=[
            httpx.Response(429, json=json_429),
            httpx.Response(429, json=json_429),
            httpx.Response(429, json=json_429),
            httpx.Response(401),
            *[httpx.Response(200, json={})] * 100,  # Just simulate OK responses from now on
        ]
    )
    with mock.patch("asyncio.sleep", new_callable=mock.AsyncMock):
        await account.get_vehicles()
    assert len(account.vehicles) == get_fingerprint_state_count()


@pytest.mark.asyncio
async def test_401_after_429_fail(bmw_fixture: respx.Router):
    """Test the error handling, when a 401 is received after exactly 3 429."""
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)

    json_429 = {"statusCode": 429, "message": "Rate limit is exceeded. Try again in 2 seconds."}

    # Fail after 3 429 and 1 401 with another 429
    bmw_fixture.get("/eadrax-vcs/v4/vehicles").mock(
        side_effect=[
            httpx.Response(429, json=json_429),
            httpx.Response(429, json=json_429),
            httpx.Response(429, json=json_429),
            httpx.Response(401),
            httpx.Response(429, json=json_429),
        ]
    )

    with mock.patch("asyncio.sleep", new_callable=mock.AsyncMock):
        with pytest.raises(MyBMWQuotaError):
            await account.get_vehicles()


@pytest.mark.asyncio
async def test_403_quota_exceeded_vehicles_usa(caplog, bmw_fixture: respx.Router):
    """Test 403 quota issues for vehicle state and fail if it happens too often."""
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)
    # get vehicles once
    await account.get_vehicles()

    bmw_fixture.get("/eadrax-vcs/v4/vehicles/state").mock(
        return_value=httpx.Response(
            403,
            json={"statusCode": 403, "message": "Out of call volume quota. Quota will be replenished in 02:12:20."},
        )
    )
    caplog.set_level(logging.DEBUG)

    with mock.patch("asyncio.sleep", new_callable=mock.AsyncMock):
        with pytest.raises(MyBMWQuotaError):
            await account.get_vehicles()

    log_quota = [r for r in caplog.records if "quota" in r.message]
    assert len(log_quota) == 1


@pytest.mark.asyncio
async def test_incomplete_vehicle_details(caplog, bmw_fixture: respx.Router):
    """Test incorrect responses for vehicle details."""
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)
    # get vehicles once
    await account.get_vehicles()

    # We need to remove the existing state route first and add it back later as otherwise our error call is never
    # matched (respx matches by order of routes and we don't replace the existing one)
    state_route = bmw_fixture.routes.pop("state")
    # JSON, but error
    bmw_fixture.get("/eadrax-vcs/v4/vehicles/state", headers={"bmw-vin": VIN_I20}).respond(
        500, json={"statusCode": 500, "message": "Something is broken."}
    )
    # No JSON
    bmw_fixture.get("/eadrax-vcs/v4/vehicles/state", headers={"bmw-vin": VIN_G26}).respond(
        500, text=load_response(RESPONSE_DIR / "auth" / "auth_error_internal_error.txt")
    )
    bmw_fixture.routes.add(state_route, "state")

    await account.get_vehicles()

    log_error = [r for r in caplog.records if "Unable to get details" in r.message]
    assert len(log_error) == 2


@pytest.mark.asyncio
async def test_no_vehicle_details(caplog, bmw_fixture: respx.Router):
    """Test raising an exception if no responses for vehicle details are received."""
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)
    await account.get_vehicles()

    bmw_fixture.get("/eadrax-vcs/v4/vehicles/state").mock(
        return_value=httpx.Response(
            500,
            json={"statusCode": 500, "message": "Something is broken."},
        )
    )
    with pytest.raises(MyBMWAPIError):
        await account.get_vehicles()

    log_error = [r for r in caplog.records if "Unable to get details" in r.message]
    assert len(log_error) == get_fingerprint_state_count()


@pytest.mark.asyncio
async def test_client_async_only(bmw_fixture: respx.Router):
    """Test that the Authentication providers only work async."""

    with httpx.Client(auth=MyBMWAuthentication(TEST_USERNAME, TEST_PASSWORD, TEST_REGION)) as client:
        with pytest.raises(RuntimeError):
            client.get("/eadrax-ucs/v1/presentation/oauth/config")

    with httpx.Client(auth=MyBMWLoginRetry()) as client:
        with pytest.raises(RuntimeError):
            client.get("/eadrax-ucs/v1/presentation/oauth/config")


@pytest.mark.asyncio
async def test_pillow_unavailable(monkeypatch: pytest.MonkeyPatch, bmw_fixture: respx.Router):
    """Test cases if Pillow is unavailable (i.e. lib is not installed with extra [china])."""

    monkeypatch.setattr("importlib.import_module", mock.Mock(side_effect=ImportError))

    # Test china (needs to throw exception)
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, get_region_from_name("china"))
    with pytest.raises(
        expected_exception=ImportError,
        match=r"Missing dependencies for region 'china'. Please install using bimmerconnected\[china\].",
    ):
        await account.get_vehicles()
    assert account is not None
    assert len(account.vehicles) == 0

    # But rest_of_world and north_america should work
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, get_region_from_name("rest_of_world"))
    await account.get_vehicles()
    assert account is not None
    assert len(account.vehicles) > 0

    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, get_region_from_name("north_america"))
    await account.get_vehicles()
    assert account is not None
    assert len(account.vehicles) > 0



"""Test for VehicleState."""

import datetime

import pytest
import respx
import time_machine

from bimmer_connected.api.regions import get_region_from_name
from bimmer_connected.vehicle.climate import Climate, ClimateActivityState
from bimmer_connected.vehicle.doors_windows import LidState, LockState
from bimmer_connected.vehicle.fuel_and_battery import ChargingState, FuelAndBattery
from bimmer_connected.vehicle.location import VehicleLocation
from bimmer_connected.vehicle.reports import CheckControlStatus, ConditionBasedServiceStatus

from . import (
    ALL_CHARGING_SETTINGS,
    VIN_F31,
    VIN_G01,
    VIN_G20,
    VIN_G26,
    VIN_G70,
    VIN_I01_NOREX,
    VIN_I01_REX,
    VIN_I20,
    get_deprecation_warning_count,
)
from .conftest import prepare_account_with_vehicles

UTC = datetime.timezone.utc


@pytest.mark.asyncio
@pytest.mark.parametrize("bmw_fixture", [[VIN_G26]], indirect=True)
async def test_generic(caplog, bmw_fixture: respx.Router):
    """Test generic attributes."""
    status = (await prepare_account_with_vehicles()).get_vehicle(VIN_G26)

    expected = datetime.datetime(year=2023, month=1, day=4, hour=14, minute=57, second=6, tzinfo=UTC)
    assert expected == status.timestamp

    assert 1121 == status.mileage[0]
    assert "km" == status.mileage[1]

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_generic_error_handling(caplog, bmw_fixture: respx.Router):
    """Test error handling when vehicle is set up."""
    account = await prepare_account_with_vehicles()

    vehicle = account.get_vehicle(VIN_G26)
    state_wo_front_left = vehicle.data["state"].copy()
    state_wo_front_left["tireState"]["frontLeft"].pop("status", None)
    vehicle.tires = None

    assert vehicle.tires is None
    vehicle.update_state(vehicle.data, state_wo_front_left)
    assert (
        any("tires" in r.message and "TypeError" in r.message and "'status'" in r.message for r in caplog.records)
        is True
    )
    assert vehicle.tires is None

    caplog.clear()

    vehicle = account.get_vehicle(VIN_I20)
    state_wo_climate_activity = vehicle.data["state"].copy()
    state_wo_climate_activity["climateControlState"].pop("activity", None)
    vehicle.climate = Climate(account_timezone=account.timezone)

    assert vehicle.climate.activity == ClimateActivityState.UNKNOWN
    vehicle.update_state(vehicle.data, state_wo_climate_activity)
    assert (
        any("climate" in r.message and "KeyError" in r.message and "'activity'" in r.message for r in caplog.records)
        is True
    )
    assert vehicle.climate.activity == ClimateActivityState.UNKNOWN


@pytest.mark.asyncio
async def test_range_combustion_no_info(caplog, bmw_fixture: respx.Router):
    """Test if the parsing of mileage and range is working."""
    vehicle = (await prepare_account_with_vehicles()).get_vehicle(VIN_F31)
    status = vehicle.fuel_and_battery

    assert (14, "L") == status.remaining_fuel
    assert status.remaining_range_fuel == (None, None)
    assert status.remaining_fuel_percent is None

    assert status.remaining_battery_percent is None
    assert status.remaining_range_electric == (None, None)

    assert status.remaining_range_total == (None, None)

    status_from_vehicle_data = FuelAndBattery.from_vehicle_data(vehicle.data)
    status_from_vehicle_data.account_timezone = status.account_timezone
    assert status_from_vehicle_data == status
    assert FuelAndBattery.from_vehicle_data({}) is None

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_range_combustion(caplog, bmw_fixture: respx.Router):
    """Test if the parsing of mileage and range is working."""
    status = (await prepare_account_with_vehicles()).get_vehicle(VIN_G20).fuel_and_battery

    assert (40, "L") == status.remaining_fuel
    assert (629, "km") == status.remaining_range_fuel
    assert status.remaining_fuel_percent == 80

    assert status.remaining_battery_percent is None
    assert status.remaining_range_electric == (None, None)

    assert (629, "km") == status.remaining_range_total

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_range_phev(caplog, bmw_fixture: respx.Router):
    """Test if the parsing of mileage and range is working."""
    status = (await prepare_account_with_vehicles()).get_vehicle(VIN_G01).fuel_and_battery

    assert (40, "L") == status.remaining_fuel
    assert (436, "km") == status.remaining_range_fuel
    assert 80 == status.remaining_fuel_percent

    assert 80 == status.remaining_battery_percent
    assert (40, "km") == status.remaining_range_electric

    assert (476, "km") == status.remaining_range_total

    assert status.remaining_range_fuel[0] + status.remaining_range_electric[0] == status.remaining_range_total[0]

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_range_rex(caplog, bmw_fixture: respx.Router):
    """Test if the parsing of mileage and range is working."""
    status = (await prepare_account_with_vehicles()).get_vehicle(VIN_I01_REX).fuel_and_battery

    assert (6, "L") == status.remaining_fuel
    assert (105, "km") == status.remaining_range_fuel
    assert status.remaining_fuel_percent is None

    assert 82 == status.remaining_battery_percent
    assert (174, "km") == status.remaining_range_electric

    assert (279, "km") == status.remaining_range_total

    assert status.remaining_range_fuel[0] + status.remaining_range_electric[0] == status.remaining_range_total[0]

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_range_electric(caplog, bmw_fixture: respx.Router):
    """Test if the parsing of mileage and range is working."""
    status = (await prepare_account_with_vehicles()).get_vehicle(VIN_I20).fuel_and_battery

    assert status.remaining_fuel == (None, None)
    assert status.remaining_range_fuel == (None, None)
    assert status.remaining_fuel_percent is None

    assert 70 == status.remaining_battery_percent
    assert (340, "km") == status.remaining_range_electric

    assert (340, "km") == status.remaining_range_total

    assert len(get_deprecation_warning_count(caplog)) == 0


@time_machine.travel("2021-11-28 21:28:59 +0000", tick=False)
@pytest.mark.asyncio
async def test_charging_end_time(caplog, bmw_fixture: respx.Router):
    """Test charging end time."""
    account = await prepare_account_with_vehicles()
    vehicle = account.get_vehicle(VIN_I01_NOREX)

    assert vehicle.fuel_and_battery.charging_end_time.astimezone(UTC) == datetime.datetime(
        2021, 11, 28, 23, 27, 59, tzinfo=UTC
    )
    assert vehicle.fuel_and_battery.charging_status == ChargingState.CHARGING
    assert vehicle.fuel_and_battery.is_charger_connected is True
    assert vehicle.fuel_and_battery.charging_start_time is None

    assert len(get_deprecation_warning_count(caplog)) == 0


@time_machine.travel("2021-11-28 17:28:59 +0000", tick=False)
@pytest.mark.asyncio
async def test_plugged_in_waiting_for_charge_window(caplog, bmw_fixture: respx.Router):
    """I01_REX is plugged in but not charging, as its waiting for charging window."""
    account = await prepare_account_with_vehicles()
    vehicle = account.get_vehicle(VIN_I01_REX)

    assert vehicle.fuel_and_battery.charging_end_time is None
    assert vehicle.fuel_and_battery.charging_status == ChargingState.WAITING_FOR_CHARGING
    assert vehicle.fuel_and_battery.is_charger_connected is True
    assert vehicle.fuel_and_battery.charging_start_time.astimezone(UTC) == datetime.datetime(
        2021, 11, 28, 18, 1, tzinfo=account.timezone
    )
    assert vehicle.fuel_and_battery.charging_target == 100

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_condition_based_services(caplog, bmw_fixture: respx.Router):
    """Test condition based service messages."""
    vehicle = (await prepare_account_with_vehicles()).get_vehicle(VIN_G26)

    cbs = vehicle.condition_based_services.messages
    assert 5 == len(cbs)
    assert ConditionBasedServiceStatus.OK == cbs[0].state
    expected_cbs0 = datetime.datetime(year=2024, month=12, day=1, tzinfo=UTC)
    assert expected_cbs0 == cbs[0].due_date
    assert (50000, "km") == cbs[0].due_distance

    assert ConditionBasedServiceStatus.OK == cbs[1].state
    expected_cbs1 = datetime.datetime(year=2024, month=12, day=1, tzinfo=UTC)
    assert expected_cbs1 == cbs[1].due_date
    assert (50000, "km") == cbs[1].due_distance

    assert ConditionBasedServiceStatus.OK == cbs[2].state
    expected_cbs2 = datetime.datetime(year=2024, month=12, day=1, tzinfo=UTC)
    assert expected_cbs2 == cbs[2].due_date
    assert (50000, "km") == cbs[2].due_distance

    assert vehicle.condition_based_services.is_service_required is False

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_position_generic(caplog, bmw_fixture: respx.Router):
    """Test generic attributes."""
    status = (await prepare_account_with_vehicles()).get_vehicle(VIN_G26)

    assert (48.177334, 11.556274) == status.vehicle_location.location
    assert 180 == status.vehicle_location.heading

    assert VehicleLocation.from_vehicle_data(status.data).location == status.vehicle_location.location

    assert VehicleLocation.from_vehicle_data({}) is None

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_vehicle_active(caplog, bmw_fixture: respx.Router):
    """Test that vehicle_active is always False."""
    account = await prepare_account_with_vehicles()

    for vehicle in account.vehicles:
        assert vehicle.is_vehicle_active is False

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_parse_f31_no_position(caplog, bmw_fixture: respx.Router):
    """Test parsing of F31 data with position tracking disabled in the vehicle."""
    vehicle = (await prepare_account_with_vehicles()).get_vehicle(VIN_F31)

    assert vehicle.vehicle_location.location is None
    assert vehicle.vehicle_location.heading is None

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_parse_gcj02_position(caplog, bmw_fixture: respx.Router):
    """Test conversion of GCJ02 to WGS84 for china."""
    account = await prepare_account_with_vehicles(get_region_from_name("china"))
    vehicle = account.get_vehicle(VIN_G01)

    vehicle_test_data = {
        "state": {
            "location": {
                "address": {"formatted": "some_formatted_address"},
                "coordinates": {"latitude": 39.83492, "longitude": 116.23221},
                "heading": 123,
            },
            "lastUpdatedAt": "2021-11-14T20:20:21Z",
        },
    }

    vehicle.update_state(dict(vehicle.data, **vehicle_test_data))

    # Update twice to test against slowly crawling position due to GCJ02 to WGS84 conversion
    vehicle.update_state(dict(vehicle.data, **vehicle_test_data))

    assert (39.8337, 116.22617) == (
        round(vehicle.vehicle_location.location[0], 5),
        round(vehicle.vehicle_location.location[1], 5),
    )

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_lids(caplog, bmw_fixture: respx.Router):
    """Test features around lids."""
    # status = (await prepare_account_with_vehicles()).get_vehicle(VIN_G01).doors_and_windows

    # assert 6 == len(list(status.lids))
    # assert 3 == len(list(status.open_lids))
    # assert status.all_lids_closed is False

    status = (await prepare_account_with_vehicles()).get_vehicle(VIN_G26).doors_and_windows

    for lid in status.lids:
        assert LidState.CLOSED == lid.state
    assert status.all_lids_closed is True
    assert 6 == len(list(status.lids))

    status = (await prepare_account_with_vehicles()).get_vehicle(VIN_I01_REX).doors_and_windows

    for lid in status.lids:
        assert LidState.CLOSED == lid.state
    assert status.all_lids_closed is True
    assert 7 == len(list(status.lids))

    assert status.lids[-1].name == "sunRoof"

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_windows_g01(caplog, bmw_fixture: respx.Router):
    """Test features around windows."""
    status = (await prepare_account_with_vehicles()).get_vehicle(VIN_G01).doors_and_windows

    for window in status.windows:
        assert LidState.CLOSED == window.state

    assert 5 == len(list(status.windows))
    assert 0 == len(list(status.open_windows))
    assert status.all_windows_closed is True

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_door_locks(caplog, bmw_fixture: respx.Router):
    """Test the door locks."""
    status = (await prepare_account_with_vehicles()).get_vehicle(VIN_G01).doors_and_windows

    assert LockState.LOCKED == status.door_lock_state

    status = (await prepare_account_with_vehicles()).get_vehicle(VIN_I01_REX).doors_and_windows

    assert LockState.UNLOCKED == status.door_lock_state

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_check_control_messages(caplog, bmw_fixture: respx.Router):
    """Test handling of check control messages.

    F11 is the only vehicle with active Check Control Messages, so we only expect to get something there.
    However we have no vehicle with issues in check control.
    """
    vehicle = (await prepare_account_with_vehicles()).get_vehicle(VIN_G01)
    assert vehicle.check_control_messages.has_check_control_messages is True

    ccms = vehicle.check_control_messages.messages
    assert 2 == len(ccms)

    assert CheckControlStatus.MEDIUM == ccms[1].state
    assert "ENGINE_OIL" == ccms[1].description_short
    assert None is ccms[1].description_long

    vehicle = (await prepare_account_with_vehicles()).get_vehicle(VIN_G20)
    assert vehicle.check_control_messages.has_check_control_messages is False

    ccms = vehicle.check_control_messages.messages
    assert 2 == len(ccms)

    assert CheckControlStatus.LOW == ccms[1].state
    assert "ENGINE_OIL" == ccms[1].description_short
    assert None is ccms[1].description_long

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_charging_profile(caplog, bmw_fixture: respx.Router):
    """Test parsing of the charging profile."""

    charging_profile = (await prepare_account_with_vehicles()).get_vehicle(VIN_I01_REX).charging_profile
    assert charging_profile.is_pre_entry_climatization_enabled is False

    departure_timer = charging_profile.departure_times[0]
    assert departure_timer.timer_id == 1
    assert departure_timer.start_time == datetime.time(7, 35)
    assert departure_timer.action == "DEACTIVATE"
    assert departure_timer.weekdays == ["MONDAY", "TUESDAY", "WEDNESDAY", "THURSDAY", "FRIDAY"]

    assert charging_profile.departure_times[3].start_time is None

    charging_window = charging_profile.preferred_charging_window
    assert charging_window.start_time == datetime.time(18, 1)
    assert charging_window.end_time == datetime.time(1, 30)

    assert charging_profile.ac_available_limits is None

    charging_settings = (await prepare_account_with_vehicles()).get_vehicle(VIN_G26).charging_profile
    assert charging_settings.ac_current_limit == 16
    assert charging_settings.ac_available_limits == [6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 20, 32]

    assert len(get_deprecation_warning_count(caplog)) == 0


@pytest.mark.asyncio
async def test_charging_profile_format_for_remote_service(caplog, bmw_fixture: respx.Router):
    """Test formatting of the charging profile."""
    account = await prepare_account_with_vehicles()

    for vin in ALL_CHARGING_SETTINGS:
        vehicle = account.get_vehicle(vin)

        fixture_data = {
            **ALL_CHARGING_SETTINGS[vin]["chargeAndClimateTimerDetail"],
            "servicePack": ALL_CHARGING_SETTINGS[vin]["servicePack"],
        }
        fixture_data["chargingMode"]["timerChange"] = "NO_CHANGE"

        # Add milliseconds to data because BMW has different formats to set & get
        fixture_data["chargingMode"]["startTimeSlot"] = f"{fixture_data['chargingMode']['startTimeSlot']}.000"
        fixture_data["chargingMode"]["endTimeSlot"] = f"{fixture_data['chargingMode']['endTimeSlot']}.000"
        for w in fixture_data["departureTimer"]["weeklyTimers"]:
            w["time"] = f"{w['time']}.000"

        assert vehicle.charging_profile.format_for_remote_service() == fixture_data


@pytest.mark.asyncio
async def test_tires(bmw_fixture: respx.Router):
    """Test tire status."""
    account = await prepare_account_with_vehicles()

    # Older vehicles do not provide tire status
    assert account.get_vehicle(VIN_F31).tires is None

    # Vehicle with tire pressure, but no details
    tires = account.get_vehicle(VIN_G01).tires
    assert tires.front_left.current_pressure == 270
    assert tires.front_left.target_pressure == 260
    assert tires.front_left.manufacturing_week is None
    assert tires.front_left.season is None

    # Vehicle with current tire pressure and details, but no target pressure
    tires = account.get_vehicle(VIN_G20).tires
    assert tires.front_left.current_pressure == 241
    assert tires.front_left.target_pressure is None
    assert tires.front_left.manufacturing_week == datetime.datetime(2021, 10, 4, 0, 0)
    assert tires.front_left.season == 2

    # Vehicle with details
    tires = account.get_vehicle(VIN_G70).tires
    assert tires.rear_left.current_pressure == 261
    assert tires.rear_left.target_pressure == 269
    assert tires.rear_left.manufacturing_week == datetime.datetime(2021, 10, 4, 0, 0)
    assert tires.rear_left.season == 2


@time_machine.travel("2021-11-28 21:28:59 +0000", tick=False)
@pytest.mark.asyncio
async def test_climate(bmw_fixture: respx.Router):
    """Test climate status."""
    account = await prepare_account_with_vehicles()

    # Older vehicles do not provide climate status
    climate = account.get_vehicle(VIN_I01_REX).climate
    assert climate.activity == ClimateActivityState.UNKNOWN
    assert climate.activity_end_time is None
    assert climate.is_climate_on is False

    # Vehicle with climate state doing nothing
    climate = account.get_vehicle(VIN_G01).climate
    assert climate.activity == ClimateActivityState.STANDBY
    assert climate.activity_end_time is None
    assert climate.is_climate_on is False

    # Running climatization
    climate = account.get_vehicle(VIN_G26).climate
    assert climate.activity == ClimateActivityState.HEATING
    assert climate.activity_end_time.astimezone(datetime.timezone.utc) == datetime.datetime(
        2021, 11, 28, 21, 58, 49, tzinfo=UTC
    )
    assert climate.is_climate_on is True



"""Generals models used for bimmer_connected."""


import datetime
import logging
from dataclasses import dataclass
from typing import Any, Dict, Optional

from bimmer_connected.const import ATTR_ATTRIBUTES, ATTR_STATE
from bimmer_connected.models import StrEnum, ValueWithUnit, VehicleDataBase
from bimmer_connected.vehicle.const import COMBUSTION_ENGINE_DRIVE_TRAINS, HV_BATTERY_DRIVE_TRAINS, DriveTrainType

_LOGGER = logging.getLogger(__name__)


class ChargingState(StrEnum):
    """Charging state of electric vehicle."""

    DEFAULT = "DEFAULT"
    CHARGING = "CHARGING"
    ERROR = "ERROR"
    COMPLETE = "COMPLETE"
    FULLY_CHARGED = "FULLY_CHARGED"
    FINISHED_FULLY_CHARGED = "FINISHED_FULLY_CHARGED"
    FINISHED_NOT_FULL = "FINISHED_NOT_FULL"
    INVALID = "INVALID"
    NOT_CHARGING = "NOT_CHARGING"
    PLUGGED_IN = "PLUGGED_IN"
    WAITING_FOR_CHARGING = "WAITING_FOR_CHARGING"
    TARGET_REACHED = "TARGET_REACHED"
    UNKNOWN = "UNKNOWN"


@dataclass
class FuelAndBattery(VehicleDataBase):
    """Provides an accessible version of `status.FuelAndBattery`."""

    remaining_range_fuel: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Get the remaining range of the vehicle on fuel."""

    remaining_range_electric: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Get the remaining range of the vehicle on electricity."""

    remaining_range_total: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Get the total remaining range of the vehicle (fuel + electricity, if available)."""

    remaining_fuel: Optional[ValueWithUnit] = ValueWithUnit(None, None)
    """Get the remaining fuel of the vehicle."""

    remaining_fuel_percent: Optional[int] = None
    """State of charge of the high voltage battery in percent."""

    remaining_battery_percent: Optional[int] = None
    """State of charge of the high voltage battery in percent."""

    charging_status: Optional[ChargingState] = None
    """Charging state of the vehicle."""

    charging_start_time_no_tz: Optional[datetime.datetime] = None
    """The planned time the vehicle will start charging without time zone information."""

    charging_end_time: Optional[datetime.datetime] = None
    """The estimated time the vehicle will have finished charging."""

    is_charger_connected: bool = False
    """Get status of the connection"""

    charging_target: Optional[int] = None
    """State of charging target in percent."""

    account_timezone: datetime.timezone = datetime.timezone.utc

    @property
    def charging_start_time(self) -> Optional[datetime.datetime]:
        """The planned time the vehicle will start charging."""
        if self.charging_start_time_no_tz:
            return self.charging_start_time_no_tz.astimezone(self.account_timezone)
        return None

    @classmethod
    def from_vehicle_data(cls, vehicle_data: Dict):
        """Create the class based on vehicle data from API."""
        parsed = cls._parse_vehicle_data(vehicle_data) or {}
        if len(parsed) > 0:
            return cls(**parsed)
        return None

    @classmethod
    def _parse_vehicle_data(cls, vehicle_data: Dict) -> Optional[Dict]:
        """Parse fuel indicators based on Ids."""
        retval: Dict[str, Any] = {}
        drivetrain = DriveTrainType(vehicle_data.get(ATTR_ATTRIBUTES, {}).get("driveTrain") or DriveTrainType.UNKNOWN)

        # return early if state data hasn't been loaded or no combustion/electric data is available
        if (
            ATTR_STATE not in vehicle_data
            or len({"combustionFuelLevel", "electricChargingState"}.intersection(set(vehicle_data[ATTR_STATE]))) == 0
        ):
            return retval

        state = vehicle_data["state"]

        if drivetrain in COMBUSTION_ENGINE_DRIVE_TRAINS:
            retval.update(cls._parse_fuel_data(state.get("combustionFuelLevel", {})))

        if drivetrain in HV_BATTERY_DRIVE_TRAINS:
            electric_data = state.get("electricChargingState", {})
            if electric_data:
                retval.update(
                    cls._parse_electric_data(
                        electric_data,
                        vehicle_data["fetched_at"],
                        state.get("chargingProfile", {}).get("reductionOfChargeCurrent"),
                    ),
                )

        if drivetrain in set(COMBUSTION_ENGINE_DRIVE_TRAINS).intersection(HV_BATTERY_DRIVE_TRAINS):
            # for hybrid vehicles the remaining_range_fuel returned by the API seems to be the total remaining range
            # to calculate the correct remaining range on fuel, we have to subtract the remaining electric range
            # this does not seem to apply for the old I3 with range extender

            fuel: ValueWithUnit = retval.get("remaining_range_fuel", ValueWithUnit(None, None))
            electric: ValueWithUnit = retval.get("remaining_range_electric", ValueWithUnit(None, None))

            if drivetrain == DriveTrainType.ELECTRIC_WITH_RANGE_EXTENDER:
                retval["remaining_range_total"] = ValueWithUnit(
                    (fuel.value or 0) + (electric.value or 0),
                    fuel.unit or electric.unit,
                )
            else:
                retval["remaining_range_total"] = retval["remaining_range_fuel"]
                retval["remaining_range_fuel"] = ValueWithUnit(
                    (fuel.value or 0) - (electric.value or 0),
                    fuel.unit or electric.unit,
                )
        elif drivetrain in COMBUSTION_ENGINE_DRIVE_TRAINS and "remaining_range_fuel" in retval:
            retval["remaining_range_total"] = retval["remaining_range_fuel"]
        elif drivetrain in HV_BATTERY_DRIVE_TRAINS and "remaining_range_electric" in retval:
            retval["remaining_range_total"] = retval["remaining_range_electric"]

        return retval

    @staticmethod
    def _parse_fuel_data(fuel_data: Dict) -> Dict:
        """Parse fuel data."""
        retval = {}
        if "remainingFuelLiters" in fuel_data:
            retval["remaining_fuel"] = ValueWithUnit(fuel_data["remainingFuelLiters"], "L")
        if "remainingFuelPercent" in fuel_data:
            retval["remaining_fuel_percent"] = fuel_data["remainingFuelPercent"]
        if "range" in fuel_data:
            retval["remaining_range_fuel"] = ValueWithUnit(fuel_data["range"], "km")
        return retval

    @staticmethod
    def _parse_electric_data(
        electric_data: Dict, fetched_at: datetime.datetime, charging_window: Optional[Dict] = None
    ) -> Dict:
        """Parse electric data."""
        retval = {}
        if "isChargerConnected" in electric_data:
            retval["is_charger_connected"] = electric_data["isChargerConnected"]
        if "chargingLevelPercent" in electric_data:
            retval["remaining_battery_percent"] = int(electric_data["chargingLevelPercent"])
        if "range" in electric_data:
            retval["remaining_range_electric"] = ValueWithUnit(electric_data["range"], "km")
        if "chargingStatus" in electric_data:
            retval["charging_status"] = ChargingState(
                electric_data["chargingStatus"] if electric_data["chargingStatus"] != "INVALID" else "NOT_CHARGING"
            )
        if "remainingChargingMinutes" in electric_data:
            retval["charging_end_time"] = fetched_at + datetime.timedelta(
                minutes=electric_data["remainingChargingMinutes"]
            )
        if "chargingTarget" in electric_data:
            retval["charging_target"] = int(electric_data["chargingTarget"])

        if retval["charging_status"] == ChargingState.WAITING_FOR_CHARGING and isinstance(charging_window, Dict):
            retval["charging_start_time_no_tz"] = datetime.datetime.combine(
                datetime.datetime.now().date(),
                datetime.time(int(charging_window["start"]["hour"]), int(charging_window["start"]["minute"])),
            )

        return retval



"""Utils for bimmer_connected.api."""

import base64
import datetime
import hashlib
import importlib
import io
import json
import logging
import mimetypes
import random
import re
import string
from typing import Dict, List, Optional, Union
from uuid import uuid4

import httpx
from Crypto.Cipher import AES
from Crypto.Hash import SHA256
from Crypto.Random import get_random_bytes
from Crypto.Util.Padding import pad

from bimmer_connected.models import AnonymizedResponse, MyBMWAPIError, MyBMWAuthError, MyBMWQuotaError

UNICODE_CHARACTER_SET = string.ascii_letters + string.digits + "-._~"
RE_VIN = re.compile(r"(?P<vin>WB[a-zA-Z0-9]{15})")
ANONYMIZED_VINS: Dict[str, str] = {}


def generate_token(length: int = 30, chars: str = UNICODE_CHARACTER_SET) -> str:
    """Generate a random token with given length and characters."""
    rand = random.SystemRandom()
    return "".join(rand.choice(chars) for _ in range(length))


def create_s256_code_challenge(code_verifier: str) -> str:
    """Create S256 code_challenge with the given code_verifier."""
    data = hashlib.sha256(code_verifier.encode("ascii")).digest()
    return base64.urlsafe_b64encode(data).rstrip(b"=").decode("UTF-8")


def get_correlation_id() -> Dict[str, str]:
    """Generate corrlation headers."""
    correlation_id = str(uuid4())
    return {
        "x-identity-provider": "gcdm",
        "x-correlation-id": correlation_id,
        "bmw-correlation-id": correlation_id,
    }


async def handle_httpstatuserror(
    ex: httpx.HTTPStatusError,
    module: str = "API",
    log_handler: Optional[logging.Logger] = None,
    dont_raise: bool = False,
) -> None:
    """Try to extract information from response and re-raise Exception."""
    _logger = log_handler or logging.getLogger(__name__)
    _level = logging.DEBUG if dont_raise else logging.ERROR

    await ex.response.aread()

    # By default we will raise a MyBMWAPIError
    _ex_to_raise = MyBMWAPIError

    # HTTP status code is 401 or 403, raise MyBMWAuthError instead
    if ex.response.status_code in [401, 403]:
        _ex_to_raise = MyBMWAuthError

    # Quota errors can either be 429 Too Many Requests or 403 Quota Exceeded (instead of 401 Forbidden)
    if ex.response.status_code == 429 or (ex.response.status_code == 403 and "quota" in ex.response.text.lower()):
        _ex_to_raise = MyBMWQuotaError

    try:
        # Try parsing the known BMW API error JSON
        _err = ex.response.json()
        _err_message = f'{type(ex).__name__}: {_err["error"]} - {_err.get("error_description", "")}'
    except (json.JSONDecodeError, KeyError):
        # If format has changed or is not JSON
        _err_message = f"{type(ex).__name__}: {ex.response.text or str(ex)}"

    _logger.log(_level, "%s due to %s", _ex_to_raise.__name__, _err_message)

    if not dont_raise:
        raise _ex_to_raise(_err_message) from ex


def anonymize_data(json_data: Union[List, Dict]) -> Union[List, Dict]:
    """Replace parts of the logfiles containing personal information."""

    replacements = {
        "lat": 12.3456,
        "latitude": 12.3456,
        "lon": 34.5678,
        "longitude": 34.5678,
        "heading": 123,
        "licensePlate": "some_license_plate",
        "name": "some_name",
        "city": "some_city",
        "street": "some_street",
        "streetNumber": "999",
        "postalCode": "some_postal_code",
        "phone": "some_phone",
        "formatted": "some_formatted_address",
        "subtitle": "some_road \u2022 duration \u2022 -- EUR",
    }

    if isinstance(json_data, list):
        json_data = [anonymize_data(v) for v in json_data]
    elif isinstance(json_data, dict):
        for key, value in json_data.items():
            if key in replacements:
                json_data[key] = replacements[key]
            elif isinstance(value, str):
                json_data[key] = RE_VIN.sub(anonymize_vin, json_data[key])
            else:
                json_data[key] = anonymize_data(value)

    return json_data


def anonymize_vin(match: re.Match):
    """Anonymize VINs but keep assignment."""
    vin = match.groupdict()["vin"]
    if vin not in ANONYMIZED_VINS:
        ANONYMIZED_VINS[vin] = f"{vin[:3]}0FINGERPRINT{str(len(ANONYMIZED_VINS) + 1).zfill(2)}"
    return ANONYMIZED_VINS[vin]


def anonymize_response(response: httpx.Response) -> AnonymizedResponse:
    """Anonymize a responses URL and content."""
    brand = response.request.headers.get("x-user-agent", ";").split(";")[1]
    brand = f"{brand}-" if brand else ""

    url_parts = response.url.path.split("/")[1:]
    if "bmw-vin" in response.request.headers:
        url_parts.append(response.request.headers["bmw-vin"])
    url_path = RE_VIN.sub(anonymize_vin, "_".join(url_parts))

    try:
        content: Union[List, Dict, str]
        content = anonymize_data(response.json())
    except json.JSONDecodeError:
        content = response.text

    content_type = next(iter((response.headers.get("content-type") or "").split(";")), "")
    file_extension = mimetypes.guess_extension(content_type or ".txt")

    return AnonymizedResponse(f"{brand}{url_path}{file_extension}", content)


def generate_random_base64_string(size: int) -> str:
    """Generate a random base64 string with size."""
    return get_random_bytes(size).hex()[:size]


def generate_cn_nonce(username: str) -> str:
    """Generate a x-login-nonce string."""
    key = generate_random_base64_string(16)
    iv = generate_random_base64_string(16)

    k1 = key[:8]
    i1 = iv[:8]
    k2 = key[8:]
    i2 = iv[8:]

    if username is None:
        username = ""

    sha256_hex = SHA256.new((k2 + i1 + "u3.6.1" + username[-4:] + k1 + i2).encode()).hexdigest()
    sha256_a = sha256_hex[:32]
    sha256_b = sha256_hex[32:]

    random_str = k1

    time_str = datetime.datetime.now(tz=datetime.timezone.utc).strftime("%a, %d %b %Y %H:%M:%S GMT")
    phone_text = f"{username}&{time_str}&{random_str}"

    chars = list("01234abcdefghijklmnopqrstuvwxyz56789")
    k1r = chars[36 - chars.index(k1[4]) - 1]
    k2r = chars[36 - chars.index(k2[4]) - 1]
    aes_key = key[:4] + k1r + key[5:12] + k2r + key[13:]

    cipher_aes = AES.new(aes_key.encode(), AES.MODE_CBC, iv.encode())

    aes_hex = cipher_aes.encrypt(pad(phone_text.encode(), AES.block_size)).hex()

    return k2 + i1 + sha256_a + aes_hex + k1 + i2 + sha256_b


def get_capture_position(base64_background_img: str) -> str:
    """Get the position of the capture in the background image."""
    target_color = [220, 230, 221]
    tolerance = 15
    block = {"width": 15, "height": 75}

    img_bytes = io.BytesIO(base64.b64decode(base64_background_img))
    img = try_import_pillow_image().open(img_bytes)
    pixels = list(img.getdata())

    position = ""
    found_block = False

    for y in range(0, img.height - block["height"]):
        for x in range(0, img.width - block["width"]):
            found_block = True
            for i in range(block["height"]):
                for j in range(block["width"]):
                    pixel_index = (y + i) * img.width + (x + j)
                    pr = pixels[pixel_index][:3]
                    dr = abs(pr[0] - target_color[0])
                    dg = abs(pr[1] - target_color[1])
                    db = abs(pr[2] - target_color[2])

                    if dr > tolerance or dg > tolerance or db > tolerance:
                        found_block = False
                        break
                if not found_block:
                    break
            if found_block:
                position = str(round((x - 26) / img.width, 2))
                break
        if found_block:
            break

    return position


def try_import_pillow_image():
    """Try to import PIL.Image and return if successful.

    We only need to load PIL if we are in China, so we try to avoid a general dependency
    on Pillow for all users. Installing Pillow on Raspberry Pi (ARMv7) is painful.
    """

    try:
        image = importlib.import_module("PIL.Image")
    except ImportError as ex:
        raise ImportError(
            "Missing dependencies for region 'china'. Please install using bimmerconnected[china]."
        ) from ex
    return image



"""Fixtures for BMW tests."""
from collections import deque
from typing import Deque, Generator, Optional

import pytest
import respx

from bimmer_connected.account import MyBMWAccount
from bimmer_connected.const import Regions
from bimmer_connected.models import AnonymizedResponse

from . import (
    ALL_CHARGING_SETTINGS,
    ALL_STATES,
    TEST_PASSWORD,
    TEST_REGION,
    TEST_USERNAME,
)
from .common import MyBMWMockRouter


@pytest.fixture
def bmw_fixture(request: pytest.FixtureRequest) -> Generator[respx.MockRouter, None, None]:
    """Patch MyBMW login API calls."""
    # Now we can start patching the API calls
    router = MyBMWMockRouter(
        vehicles_to_load=getattr(request, "param", []),
        states=ALL_STATES,
        charging_settings=ALL_CHARGING_SETTINGS,
    )

    with router:
        yield router


@pytest.fixture
def bmw_log_all_responses(monkeypatch: pytest.MonkeyPatch):
    """Increase the length of the response store to log all responses."""
    temp_store: Deque[AnonymizedResponse] = deque(maxlen=100)
    monkeypatch.setattr("bimmer_connected.api.client.RESPONSE_STORE", temp_store)
    monkeypatch.setattr("bimmer_connected.account.RESPONSE_STORE", temp_store)


async def prepare_account_with_vehicles(region: Optional[Regions] = None):
    """Initialize account and get vehicles."""
    account = MyBMWAccount(TEST_USERNAME, TEST_PASSWORD, region or TEST_REGION)
    await account.get_vehicles()
    return account



"""Fixtures for BMW tests."""

import json
from collections import defaultdict
from copy import deepcopy
from pathlib import Path
from typing import Dict, List, Optional
from unittest import mock
from uuid import uuid4

import httpx
import respx

from bimmer_connected.const import Regions
from bimmer_connected.models import ChargingSettings
from bimmer_connected.vehicle.climate import ClimateActivityState
from bimmer_connected.vehicle.fuel_and_battery import ChargingState
from bimmer_connected.vehicle.remote_services import MAP_CHARGING_MODE_TO_REMOTE_SERVICE

from . import (
    ALL_VEHICLES,
    REMOTE_SERVICE_RESPONSE_DELIVERED,
    REMOTE_SERVICE_RESPONSE_EVENTPOSITION,
    REMOTE_SERVICE_RESPONSE_EXECUTED,
    REMOTE_SERVICE_RESPONSE_INITIATED,
    REMOTE_SERVICE_RESPONSE_PENDING,
    RESPONSE_DIR,
    load_response,
)

POI_DATA = {
    "lat": 37.4028943,
    "lon": -121.9700289,
    "name": "49ers",
    "street": "4949 Marie P DeBartolo Way",
    "city": "Santa Clara",
    "postal_code": "CA 95054",
    "country": "United States",
}

CHARGING_SETTINGS = {"target_soc": 75, "ac_limit": 16}

STATUSREMOTE_SERVICE_RESPONSE_ORDER = [
    REMOTE_SERVICE_RESPONSE_PENDING,
    REMOTE_SERVICE_RESPONSE_DELIVERED,
    REMOTE_SERVICE_RESPONSE_EXECUTED,
]
STATUSREMOTE_SERVICE_RESPONSE_DICT: Dict[str, List[Path]] = defaultdict(
    lambda: deepcopy(STATUSREMOTE_SERVICE_RESPONSE_ORDER)
)

MAP_REMOTE_SERVICE_CHARGING_MODE_TO_STATE = {v: k.value for k, v in MAP_CHARGING_MODE_TO_REMOTE_SERVICE.items()}

LOCAL_STATES: Dict[str, Dict] = {}
LOCAL_CHARGING_SETTINGS: Dict[str, Dict] = {}


class MyBMWMockRouter(respx.MockRouter):
    """Stateful MockRouter for MyBMW APIs."""

    def __init__(
        self,
        vehicles_to_load: Optional[List[str]] = None,
        states: Optional[Dict[str, Dict]] = None,
        charging_settings: Optional[Dict[str, Dict]] = None,
    ) -> None:
        """Initialize the MyBMWMockRouter with clean responses."""
        super().__init__(assert_all_called=False)
        self.vehicles_to_load = vehicles_to_load or []
        self.states = deepcopy(states) if states else {}
        self.charging_settings = deepcopy(charging_settings) if charging_settings else {}

        self.add_login_routes()
        self.add_vehicle_routes()
        self.add_remote_service_routes()

    # # # # # # # # # # # # # # # # # # # # # # # #
    # Routes
    # # # # # # # # # # # # # # # # # # # # # # # #

    def add_login_routes(self) -> None:
        """Add routes for login."""

        # Login to north_america and rest_of_world
        self.get("/eadrax-ucs/v1/presentation/oauth/config").respond(
            200, json=load_response(RESPONSE_DIR / "auth" / "oauth_config.json")
        )
        self.post("/gcdm/oauth/authenticate").mock(side_effect=self.authenticate_sideeffect)
        self.post("/gcdm/oauth/token").respond(200, json=load_response(RESPONSE_DIR / "auth" / "auth_token.json"))

        # Login to china
        self.get("/eadrax-coas/v1/cop/publickey").respond(
            200, json=load_response(RESPONSE_DIR / "auth" / "auth_cn_publickey.json")
        )
        self.post("/eadrax-coas/v2/cop/slider-captcha").respond(
            200, json=load_response(RESPONSE_DIR / "auth" / "auth_slider_captcha.json")
        )

        self.post("/eadrax-coas/v1/cop/check-captcha").respond(
            200, json=load_response(RESPONSE_DIR / "auth" / "auth_slider_captcha_check.json")
        )

        self.post("/eadrax-coas/v2/login/pwd").respond(
            200, json=load_response(RESPONSE_DIR / "auth" / "auth_cn_login_pwd.json")
        )
        self.post("/eadrax-coas/v2/oauth/token").respond(
            200, json=load_response(RESPONSE_DIR / "auth" / "auth_token.json")
        )

    def add_vehicle_routes(self) -> None:
        """Add routes for vehicle requests."""

        self.get("/eadrax-vcs/v4/vehicles").mock(side_effect=self.vehicles_sideeffect)
        self.get("/eadrax-vcs/v4/vehicles/state", name="state").mock(side_effect=self.vehicle_state_sideeffect)
        self.get("/eadrax-crccs/v2/vehicles").mock(side_effect=self.vehicle_charging_settings_sideeffect)

    def add_remote_service_routes(self) -> None:
        """Add routes for remote services."""

        self.post(path__regex=r"/eadrax-vrccs/v3/presentation/remote-commands/(?P<vin>.+)/(?P<service>.+)$").mock(
            side_effect=self.service_trigger_sideeffect
        )
        self.post(path__regex=r"/eadrax-crccs/v1/vehicles/(?P<vin>.+)/(?P<service>(start|stop)-charging)$").mock(
            side_effect=self.service_trigger_sideeffect
        )
        self.post(path__regex=r"/eadrax-crccs/v1/vehicles/(?P<vin>.+)/charging-settings$").mock(
            side_effect=self.charging_settings_sideeffect
        )
        self.post(path__regex=r"/eadrax-crccs/v1/vehicles/(?P<vin>.+)/charging-profile$").mock(
            side_effect=self.charging_profile_sideeffect
        )
        self.post("/eadrax-vrccs/v3/presentation/remote-commands/eventStatus", params={"eventId": mock.ANY}).mock(
            side_effect=self.service_status_sideeffect
        )

        self.post("/eadrax-dcs/v1/send-to-car/send-to-car").mock(side_effect=self.poi_sideeffect)
        self.post("/eadrax-vrccs/v3/presentation/remote-commands/eventPosition", params={"eventId": mock.ANY}).respond(
            200,
            json=load_response(REMOTE_SERVICE_RESPONSE_EVENTPOSITION),
        )

    # # # # # # # # # # # # # # # # # # # # # # # #
    # Authentication sideeffects
    # # # # # # # # # # # # # # # # # # # # # # # #

    @staticmethod
    def authenticate_sideeffect(request: httpx.Request) -> httpx.Response:
        """Return /oauth/authentication response based on request."""
        request_text = request.read().decode("UTF-8")
        if "username" in request_text and "password" in request_text and "grant_type" in request_text:
            return httpx.Response(200, json=load_response(RESPONSE_DIR / "auth" / "authorization_response.json"))
        return httpx.Response(
            302,
            headers={
                "Location": "com.mini.connected://oauth?code=CODE&state=STATE&client_id=CLIENT_ID&nonce=login_nonce",
            },
        )

    # # # # # # # # # # # # # # # # # # # # # # # #
    # Vehicle state sideeffects
    # # # # # # # # # # # # # # # # # # # # # # # #

    def vehicles_sideeffect(self, request: httpx.Request) -> httpx.Response:
        """Return /vehicles response based on x-user-agent."""
        x_user_agent = request.headers.get("x-user-agent", "").split(";")
        if len(x_user_agent) == 4:
            brand = x_user_agent[1]
        else:
            raise ValueError("x-user-agent not configured correctly!")

        # Test if given region is valid
        _ = Regions(x_user_agent[3])

        fingerprints = ALL_VEHICLES.get(brand, [])
        if self.vehicles_to_load:
            fingerprints = [f for f in fingerprints if f["vin"] in self.vehicles_to_load]

        # Ensure order
        fingerprints = sorted(fingerprints, key=lambda v: v["vin"])

        return httpx.Response(200, json=fingerprints)

    def vehicle_state_sideeffect(self, request: httpx.Request) -> httpx.Response:
        """Return /vehicles response based on x-user-agent."""
        x_user_agent = request.headers.get("x-user-agent", "").split(";")
        assert len(x_user_agent) == 4

        try:
            return httpx.Response(200, json=self.states[request.headers["bmw-vin"]])
        except KeyError:
            return httpx.Response(404)

    def vehicle_charging_settings_sideeffect(self, request: httpx.Request) -> httpx.Response:
        """Return /vehicles response based on x-user-agent."""
        x_user_agent = request.headers.get("x-user-agent", "").split(";")
        assert len(x_user_agent) == 4
        assert "fields" in request.url.params
        assert "has_charging_settings_capabilities" in request.url.params

        try:
            return httpx.Response(200, json=self.charging_settings[request.headers["bmw-vin"]])
        except KeyError:
            return httpx.Response(404)

    # # # # # # # # # # # # # # # # # # # # # # # #
    # Remote service sideeffects
    # # # # # # # # # # # # # # # # # # # # # # # #

    def service_trigger_sideeffect(
        self, request: httpx.Request, vin: str, service: Optional[str] = None
    ) -> httpx.Response:
        """Return specific eventId for each remote function."""

        if service in ["door-lock", "door-unlock"]:
            new_state = "LOCKED" if service == "door-lock" else "UNLOCKED"
            self.states[vin]["state"]["doorsState"]["combinedSecurityState"] = new_state

        elif service in ["climate-now"]:
            new_state = (
                ClimateActivityState.COOLING
                if request.url.params["action"] == "START"
                else ClimateActivityState.STANDBY
            )
            self.states[vin]["state"]["climateControlState"]["activity"] = new_state

        elif service in ["start-charging", "stop-charging"]:
            new_state = ChargingState.PLUGGED_IN if "stop" in service else ChargingState.CHARGING
            self.states[vin]["state"]["electricChargingState"]["chargingStatus"] = new_state

        elif service in ["vehicle-finder"]:
            # nothing to do here as this is handled by the
            # return of REMOTE_SERVICE_RESPONSE_EVENTPOSITION
            pass

        json_data = load_response(REMOTE_SERVICE_RESPONSE_INITIATED)
        json_data["eventId"] = str(uuid4())

        return httpx.Response(200, json=json_data)

    def charging_settings_sideeffect(self, request: httpx.Request, vin: str) -> httpx.Response:
        """Check if payload is a valid charging settings payload and return evendId."""
        cs = ChargingSettings(**json.loads(request.content))

        # this endpoint allows fields to be omitted, so we have to check for that
        if cs.chargingTarget:
            self.states[vin]["state"]["electricChargingState"]["chargingTarget"] = cs.chargingTarget
        if cs.acLimitValue:
            self.states[vin]["state"]["chargingProfile"]["chargingSettings"]["acCurrentLimit"] = cs.acLimitValue

        return self.service_trigger_sideeffect(request, vin)

    def charging_profile_sideeffect(self, request: httpx.Request, vin: str) -> httpx.Response:
        """Check if payload is a valid charging settings payload and return evendId."""

        data = json.loads(request.content)

        if {"chargingMode", "departureTimer", "isPreconditionForDepartureActive", "servicePack"} != set(data):
            return httpx.Response(500)
        if (
            data["chargingMode"]["chargingPreference"] == "NO_PRESELECTION"
            and data["chargingMode"]["type"] != "CHARGING_IMMEDIATELY"
        ):
            return httpx.Response(500)
        if (
            data["chargingMode"]["chargingPreference"] == "CHARGING_WINDOW"
            and data["chargingMode"]["type"] != "TIME_SLOT"
        ):
            return httpx.Response(500)

        if not isinstance(data["isPreconditionForDepartureActive"], bool):
            return httpx.Response(500)

        # this endpoint does not allow fields to be omitted, so we can assume that they are always present

        # separate charging profile endpoint
        self.charging_settings[vin]["chargeAndClimateTimerDetail"]["chargingMode"]["chargingPreference"] = data[
            "chargingMode"
        ]["chargingPreference"]
        self.charging_settings[vin]["chargeAndClimateTimerDetail"]["chargingMode"]["type"] = data["chargingMode"][
            "type"
        ]
        self.charging_settings[vin]["chargeAndClimateTimerDetail"]["isPreconditionForDepartureActive"] = data[
            "isPreconditionForDepartureActive"
        ]

        # state endpoint
        self.states[vin]["state"]["chargingProfile"]["chargingPreference"] = data["chargingMode"]["chargingPreference"]
        self.states[vin]["state"]["chargingProfile"]["chargingMode"] = MAP_REMOTE_SERVICE_CHARGING_MODE_TO_STATE[
            data["chargingMode"]["type"]
        ]
        self.states[vin]["state"]["chargingProfile"]["climatisationOn"] = data["isPreconditionForDepartureActive"]

        return self.service_trigger_sideeffect(request, vin)

    @staticmethod
    def service_status_sideeffect(request: httpx.Request) -> httpx.Response:
        """Return all 3 eventStatus responses per function."""
        response_data = STATUSREMOTE_SERVICE_RESPONSE_DICT[request.url.params["eventId"]].pop(0)
        return httpx.Response(200, json=load_response(response_data))

    @staticmethod
    def poi_sideeffect(request: httpx.Request) -> httpx.Response:
        """Check if payload is a valid POI."""
        data = json.loads(request.content)
        tests = all(
            [
                len(data["vin"]) == 17,
                isinstance(data["location"]["coordinates"]["latitude"], float),
                isinstance(data["location"]["coordinates"]["longitude"], float),
                len(data["location"]["name"]) > 0,
            ]
        )
        if not tests:
            return httpx.Response(400)
        return httpx.Response(201)



"""Authentication management for BMW APIs."""

import asyncio
import base64
import datetime
import logging
import math
from collections import defaultdict
from typing import AsyncGenerator, Generator, Optional
from uuid import uuid4

import httpx
import jwt
from Crypto.Cipher import PKCS1_v1_5
from Crypto.PublicKey import RSA

from bimmer_connected.api.regions import Regions, get_app_version, get_ocp_apim_key, get_server_url, get_user_agent
from bimmer_connected.api.utils import (
    create_s256_code_challenge,
    generate_cn_nonce,
    generate_token,
    get_capture_position,
    get_correlation_id,
    handle_httpstatuserror,
    try_import_pillow_image,
)
from bimmer_connected.const import (
    AUTH_CHINA_CAPTCHA_CHECK_URL,
    AUTH_CHINA_CAPTCHA_URL,
    AUTH_CHINA_LOGIN_URL,
    AUTH_CHINA_PUBLIC_KEY_URL,
    AUTH_CHINA_TOKEN_URL,
    HTTPX_TIMEOUT,
    OAUTH_CONFIG_URL,
    X_USER_AGENT,
)
from bimmer_connected.models import MyBMWAPIError

EXPIRES_AT_OFFSET = datetime.timedelta(seconds=HTTPX_TIMEOUT * 2)

_LOGGER = logging.getLogger(__name__)


class MyBMWAuthentication(httpx.Auth):
    """Authentication and Retry Handler for MyBMW API."""

    def __init__(
        self,
        username: str,
        password: str,
        region: Regions,
        access_token: Optional[str] = None,
        expires_at: Optional[datetime.datetime] = None,
        refresh_token: Optional[str] = None,
        gcid: Optional[str] = None,
    ):
        self.username: str = username
        self.password: str = password
        self.region: Regions = region
        self.access_token: Optional[str] = access_token
        self.expires_at: Optional[datetime.datetime] = expires_at
        self.refresh_token: Optional[str] = refresh_token
        self.session_id: str = str(uuid4())
        self._lock: Optional[asyncio.Lock] = None
        self.gcid: Optional[str] = gcid

    @property
    def login_lock(self) -> asyncio.Lock:
        """Make sure that there is a lock in the current event loop."""
        if not self._lock:
            self._lock = asyncio.Lock()
        return self._lock

    def sync_auth_flow(self, request: httpx.Request) -> Generator[httpx.Request, httpx.Response, None]:
        raise RuntimeError("Cannot use a async authentication class with httpx.Client")

    async def async_auth_flow(self, request: httpx.Request) -> AsyncGenerator[httpx.Request, httpx.Response]:
        # Get an access token on first call
        async with self.login_lock:
            if not self.access_token:
                await self.login()
        request.headers["authorization"] = f"Bearer {self.access_token}"
        request.headers["bmw-session-id"] = self.session_id

        # Try getting a response
        response: httpx.Response = (yield request)

        # return directly if first response was successful
        if response.is_success:
            return

        await response.aread()

        # First check against 429 Too Many Requests and 403 Quota Exceeded
        retry_count = 0
        while (
            response.status_code == 429 or (response.status_code == 403 and "quota" in response.text.lower())
        ) and retry_count < 3:
            # Quota errors can either be 429 Too Many Requests or 403 Quota Exceeded (instead of 403 Forbidden)
            wait_time = get_retry_wait_time(response)
            _LOGGER.debug("Sleeping %s seconds due to 429 Too Many Requests", wait_time)
            await asyncio.sleep(wait_time)
            response = yield request
            await response.aread()
            retry_count += 1

        # Handle 401 Unauthorized and try getting a new token
        if response.status_code == 401:
            async with self.login_lock:
                _LOGGER.debug("Received unauthorized response, refreshing token.")
                await self.login()
            request.headers["authorization"] = f"Bearer {self.access_token}"
            request.headers["bmw-session-id"] = self.session_id
            response = yield request

        # Raise if request still was not successful
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as ex:
            await handle_httpstatuserror(ex, log_handler=_LOGGER)

    async def login(self) -> None:
        """Get a valid OAuth token."""
        token_data = {}
        if self.region in [Regions.NORTH_AMERICA, Regions.REST_OF_WORLD]:
            # Try logging in with refresh token first
            if self.refresh_token:
                token_data = await self._refresh_token_row_na()
            if not token_data:
                token_data = await self._login_row_na()
            token_data["expires_at"] = token_data["expires_at"] - EXPIRES_AT_OFFSET

        elif self.region in [Regions.CHINA]:
            # Try logging in with refresh token first
            if self.refresh_token:
                token_data = await self._refresh_token_china()
            if not token_data:
                token_data = await self._login_china()
            token_data["expires_at"] = token_data["expires_at"] - EXPIRES_AT_OFFSET
            self.gcid = token_data["gcid"]

        self.access_token = token_data["access_token"]
        self.expires_at = token_data["expires_at"]
        self.refresh_token = token_data["refresh_token"]

    async def _login_row_na(self):
        """Login to Rest of World and North America."""
        async with MyBMWLoginClient(region=self.region) as client:
            _LOGGER.debug("Authenticating with MyBMW flow for North America & Rest of World.")

            # Get OAuth2 settings from BMW API
            r_oauth_settings = await client.get(
                OAUTH_CONFIG_URL,
                headers={
                    "ocp-apim-subscription-key": get_ocp_apim_key(self.region),
                    "bmw-session-id": self.session_id,
                    **get_correlation_id(),
                },
            )
            oauth_settings = r_oauth_settings.json()

            # Generate OAuth2 Code Challenge + State
            code_verifier = generate_token(86)
            code_challenge = create_s256_code_challenge(code_verifier)

            state = generate_token(22)

            # Set up authenticate endpoint
            authenticate_url = oauth_settings["tokenEndpoint"].replace("/token", "/authenticate")
            oauth_base_values = {
                "client_id": oauth_settings["clientId"],
                "response_type": "code",
                "redirect_uri": oauth_settings["returnUrl"],
                "state": state,
                "nonce": "login_nonce",
                "scope": " ".join(oauth_settings["scopes"]),
                "code_challenge": code_challenge,
                "code_challenge_method": "S256",
            }

            # Call authenticate endpoint first time (with user/pw) and get authentication
            response = await client.post(
                authenticate_url,
                data=dict(
                    oauth_base_values,
                    **{
                        "grant_type": "authorization_code",
                        "username": self.username,
                        "password": self.password,
                    },
                ),
            )
            authorization = httpx.URL(response.json()["redirect_to"]).params["authorization"]

            # With authorization, call authenticate endpoint second time to get code
            response = await client.post(
                authenticate_url,
                params={
                    "interaction-id": uuid4(),
                    "client-version": X_USER_AGENT.format(
                        brand="bmw", app_version=get_app_version(self.region), region=self.region.value
                    ),
                },
                data=dict(oauth_base_values, **{"authorization": authorization}),
            )
            code = response.next_request.url.params["code"]

            # With code, get token
            current_utc_time = datetime.datetime.now(tz=datetime.timezone.utc)
            response = await client.post(
                oauth_settings["tokenEndpoint"],
                data={
                    "code": code,
                    "code_verifier": code_verifier,
                    "redirect_uri": oauth_settings["returnUrl"],
                    "grant_type": "authorization_code",
                },
                auth=(oauth_settings["clientId"], oauth_settings["clientSecret"]),
            )
            response_json = response.json()

            expiration_time = int(response_json["expires_in"])
            expires_at = current_utc_time + datetime.timedelta(seconds=expiration_time)

        return {
            "access_token": response_json["access_token"],
            "expires_at": expires_at,
            "refresh_token": response_json["refresh_token"],
        }

    async def _refresh_token_row_na(self):
        """Login to Rest of World and North America using existing refresh_token."""
        try:
            async with MyBMWLoginClient(region=self.region) as client:
                _LOGGER.debug("Authenticating with refresh token for North America & Rest of World.")

                # Get OAuth2 settings from BMW API
                r_oauth_settings = await client.get(
                    OAUTH_CONFIG_URL,
                    headers={
                        "ocp-apim-subscription-key": get_ocp_apim_key(self.region),
                        "bmw-session-id": self.session_id,
                        **get_correlation_id(),
                    },
                )
                oauth_settings = r_oauth_settings.json()

                # With code, get token
                current_utc_time = datetime.datetime.now(tz=datetime.timezone.utc)
                response = await client.post(
                    oauth_settings["tokenEndpoint"],
                    data={
                        "scope": " ".join(oauth_settings["scopes"]),
                        "redirect_uri": oauth_settings["returnUrl"],
                        "grant_type": "refresh_token",
                        "refresh_token": self.refresh_token,
                    },
                    auth=(oauth_settings["clientId"], oauth_settings["clientSecret"]),
                )
                response_json = response.json()

                expiration_time = int(response_json["expires_in"])
                expires_at = current_utc_time + datetime.timedelta(seconds=expiration_time)

        except MyBMWAPIError:
            _LOGGER.debug("Unable to get access token using refresh token, falling back to username/password.")
            return {}

        return {
            "access_token": response_json["access_token"],
            "expires_at": expires_at,
            "refresh_token": response_json["refresh_token"],
        }

    async def _login_china(self):
        async with MyBMWLoginClient(region=self.region) as client:
            _LOGGER.debug("Authenticating with MyBMW flow for China.")

            # While PIL.Image is only needed in `get_capture_position`, we test it here to avoid
            # unneeded requests to the server.
            try_import_pillow_image()

            # Get current RSA public certificate & use it to encrypt password
            response = await client.get(
                AUTH_CHINA_PUBLIC_KEY_URL,
            )
            pem_public_key = response.json()["data"]["value"]

            public_key = RSA.import_key(pem_public_key)
            cipher_rsa = PKCS1_v1_5.new(public_key)
            encrypted = cipher_rsa.encrypt(self.password.encode())
            pw_encrypted = base64.b64encode(encrypted).decode("UTF-8")

            captcha_res = await client.post(
                AUTH_CHINA_CAPTCHA_URL,
                json={"mobile": self.username},
            )
            verify_id = captcha_res.json()["data"]["verifyId"]

            position = get_capture_position(captcha_res.json()["data"]["backGroundImg"])
            await client.post(AUTH_CHINA_CAPTCHA_CHECK_URL, json={"position": position, "verifyId": verify_id})

            # Get token
            response = await client.post(
                AUTH_CHINA_LOGIN_URL,
                headers={"x-login-nonce": generate_cn_nonce(self.username)},
                json={
                    "mobile": self.username,
                    "password": pw_encrypted,
                    "verifyId": verify_id,
                    "deviceId": self.username,
                },
            )
            response_json = response.json()["data"]

            decoded_token = jwt.decode(
                response_json["access_token"], algorithms=["HS256"], options={"verify_signature": False}
            )

        return {
            "access_token": response_json["access_token"],
            "expires_at": datetime.datetime.fromtimestamp(decoded_token["exp"], tz=datetime.timezone.utc),
            "refresh_token": response_json["refresh_token"],
            "gcid": response_json["gcid"],
        }

    async def _refresh_token_china(self):
        try:
            async with MyBMWLoginClient(region=self.region) as client:
                _LOGGER.debug("Authenticating with refresh token for China.")

                current_utc_time = datetime.datetime.now(tz=datetime.timezone.utc)

                # Try logging in using refresh_token
                response = await client.post(
                    AUTH_CHINA_TOKEN_URL,
                    headers={"x-login-nonce": generate_cn_nonce(self.gcid)},
                    data={
                        "refresh_token": self.refresh_token,
                        "grant_type": "refresh_token",
                    },
                )
                response_json = response.json()

                expiration_time = int(response_json["expires_in"])
                expires_at = current_utc_time + datetime.timedelta(seconds=expiration_time)

        except MyBMWAPIError:
            _LOGGER.debug("Unable to get access token using refresh token, falling back to username/password.")
            return {}

        return {
            "access_token": response_json["access_token"],
            "expires_at": expires_at,
            "refresh_token": response_json["refresh_token"],
            "gcid": response_json["gcid"],
        }


class MyBMWLoginClient(httpx.AsyncClient):
    """Async HTTP client based on `httpx.AsyncClient` with automated OAuth token refresh."""

    def __init__(self, *args, **kwargs):
        # Increase timeout
        kwargs["timeout"] = httpx.Timeout(HTTPX_TIMEOUT)

        kwargs["auth"] = MyBMWLoginRetry()

        # Set default values#
        region = kwargs.pop("region")
        kwargs["base_url"] = get_server_url(region)
        kwargs["headers"] = {
            "user-agent": get_user_agent(region),
            "x-user-agent": X_USER_AGENT.format(brand="bmw", app_version=get_app_version(region), region=region.value),
        }

        # Register event hooks
        kwargs["event_hooks"] = defaultdict(list, **kwargs.get("event_hooks", {}))

        # Event hook which calls raise_for_status on all requests
        async def raise_for_status_event_handler(response: httpx.Response):
            """Event handler that automatically raises HTTPStatusErrors when attached.

            Will only raise on 4xx/5xx errors but not 429 which is handled `self.auth`.
            """
            if response.is_error and response.status_code != 429:
                try:
                    response.raise_for_status()
                except httpx.HTTPStatusError as ex:
                    await handle_httpstatuserror(ex, log_handler=_LOGGER)

        kwargs["event_hooks"]["response"].append(raise_for_status_event_handler)

        super().__init__(*args, **kwargs)


class MyBMWLoginRetry(httpx.Auth):
    """httpx.Auth used as workaround to retry & sleep on 429 Too Many Requests."""

    def sync_auth_flow(self, request: httpx.Request) -> Generator[httpx.Request, httpx.Response, None]:
        raise RuntimeError("Cannot use a async authentication class with httpx.Client")

    async def async_auth_flow(self, request: httpx.Request) -> AsyncGenerator[httpx.Request, httpx.Response]:
        # Try getting a response
        response: httpx.Response = (yield request)

        for _ in range(3):
            if response.status_code == 429:
                await response.aread()
                wait_time = get_retry_wait_time(response)
                _LOGGER.debug("Sleeping %s seconds due to 429 Too Many Requests", wait_time)
                await asyncio.sleep(wait_time)
                response = yield request
        # Only checking for 429 errors, as all other errors are handled by the
        # response hook of MyBMWLoginClient
        if response.status_code == 429:
            try:
                response.raise_for_status()
            except httpx.HTTPStatusError as ex:
                await handle_httpstatuserror(ex, log_handler=_LOGGER)


def get_retry_wait_time(response: httpx.Response) -> int:
    """Get the wait time for the next retry from the response and multiply by 2."""
    try:
        response_wait_time = next(iter([int(i) for i in response.json().get("message", "") if i.isdigit()]))
    except Exception:
        response_wait_time = 2
    wait_time = math.ceil(response_wait_time * 2)
    return wait_time



"""Mock for Connected Drive Backend."""

import json
from pathlib import Path
from typing import Any, Dict, List, Union

from bimmer_connected.api.regions import Regions
from bimmer_connected.const import CarBrands

RESPONSE_DIR = Path(__file__).parent / "responses"

TEST_USERNAME = "some_user"
TEST_PASSWORD = "my_secret"
TEST_REGION = Regions.REST_OF_WORLD
TEST_REGION_STRING = "rest_of_world"

VIN_F31 = "WBA00000000000F31"
VIN_G01 = "WBA00000000DEMO04"
VIN_G20 = "WBA00000000DEMO03"
VIN_G26 = "WBA00000000DEMO02"
VIN_G70 = "WBA00000000DEMO05"
VIN_I01_NOREX = "WBY000000NOREXI01"
VIN_I01_REX = "WBY00000000REXI01"
VIN_I20 = "WBA00000000DEMO01"

ALL_VEHICLES: Dict[str, List[Dict]] = {brand.value: [] for brand in CarBrands}
ALL_STATES: Dict[str, Dict] = {}
ALL_CHARGING_SETTINGS: Dict[str, Dict] = {}

REMOTE_SERVICE_RESPONSE_INITIATED = RESPONSE_DIR / "remote_services" / "eadrax_service_initiated.json"
REMOTE_SERVICE_RESPONSE_PENDING = RESPONSE_DIR / "remote_services" / "eadrax_service_pending.json"
REMOTE_SERVICE_RESPONSE_DELIVERED = RESPONSE_DIR / "remote_services" / "eadrax_service_delivered.json"
REMOTE_SERVICE_RESPONSE_EXECUTED = RESPONSE_DIR / "remote_services" / "eadrax_service_executed.json"
REMOTE_SERVICE_RESPONSE_ERROR = RESPONSE_DIR / "remote_services" / "eadrax_service_error.json"
REMOTE_SERVICE_RESPONSE_EVENTPOSITION = RESPONSE_DIR / "remote_services" / "eadrax_service_eventposition.json"


def get_fingerprint_state_count() -> int:
    """Return number of loaded vehicles."""
    return sum([len(vehicles) for vehicles in ALL_VEHICLES.values()])


def get_fingerprint_charging_settings_count() -> int:
    """Return number of loaded vehicles."""
    return len(ALL_CHARGING_SETTINGS)


def load_response(path: Union[Path, str]) -> Any:
    """Load a stored response."""
    with open(path, "rb") as file:
        if Path(path).suffix == ".json":
            return json.load(file)
        return file.read().decode("UTF-8")


for fingerprint in RESPONSE_DIR.rglob("*-eadrax-vcs_v4_vehicles.json"):
    brand = fingerprint.stem.split("-")[0]
    for vehicle in load_response(fingerprint):
        ALL_VEHICLES[brand].append(vehicle)

for state in RESPONSE_DIR.rglob("*-eadrax-vcs_v4_vehicles_state_*.json"):
    ALL_STATES[state.stem.split("_")[-1]] = load_response(state)

for charging_setting in RESPONSE_DIR.rglob("*-eadrax-crccs_v2_vehicles_*.json"):
    ALL_CHARGING_SETTINGS[charging_setting.stem.split("_")[-1]] = load_response(charging_setting)


def get_deprecation_warning_count(caplog):
    """Return all logged DeprecationWarnings."""
    return [r for r in caplog.records if r.levelname == "WARNING" and "DeprecationWarning" in r.message]



#!/usr/bin/python3
"""Simple executable to demonstrate and test the usage of the library."""

import argparse
import asyncio
import json
import logging
import sys
import time
from pathlib import Path

import httpx

from bimmer_connected.account import MyBMWAccount
from bimmer_connected.api.regions import get_region_from_name, valid_regions
from bimmer_connected.const import DEFAULT_POI_NAME
from bimmer_connected.utils import MyBMWJSONEncoder, log_response_store_to_file
from bimmer_connected.vehicle import MyBMWVehicle, VehicleViewDirection
from bimmer_connected.vehicle.charging_profile import ChargingMode

TEXT_VIN = "Vehicle Identification Number"


def main_parser() -> argparse.ArgumentParser:
    """Create the ArgumentParser with all relevant subparsers."""
    logging.basicConfig(level=logging.DEBUG)

    parser = argparse.ArgumentParser(description="A simple executable to use and test the library.")
    subparsers = parser.add_subparsers(dest="cmd")
    subparsers.required = True

    status_parser = subparsers.add_parser("status", description="Get the current status of the vehicle.")
    status_parser.add_argument(
        "-j", "--json", help="Output as JSON only. Removes all other output.", action="store_true"
    )
    _add_default_arguments(status_parser)
    _add_position_arguments(status_parser)

    fingerprint_parser = subparsers.add_parser("fingerprint", description="Save a vehicle fingerprint.")
    _add_default_arguments(fingerprint_parser)
    _add_position_arguments(fingerprint_parser)
    fingerprint_parser.set_defaults(func=fingerprint)

    flash_parser = subparsers.add_parser("lightflash", description="Flash the vehicle lights.")
    _add_default_arguments(flash_parser)
    flash_parser.add_argument("vin", help=TEXT_VIN)
    flash_parser.set_defaults(func=light_flash)

    horn_parser = subparsers.add_parser("horn", description="Trigger the vehicle horn")
    _add_default_arguments(horn_parser)
    horn_parser.add_argument("vin", help=TEXT_VIN)
    horn_parser.set_defaults(func=horn)

    finder_parser = subparsers.add_parser("vehiclefinder", description="Update the vehicle GPS location.")
    _add_default_arguments(finder_parser)
    finder_parser.add_argument("vin", help=TEXT_VIN)
    _add_position_arguments(finder_parser)
    finder_parser.set_defaults(func=vehicle_finder)

    chargingsettings_parser = subparsers.add_parser("chargingsettings", description="Set vehicle charging settings.")
    _add_default_arguments(chargingsettings_parser)
    chargingsettings_parser.add_argument("vin", help=TEXT_VIN)
    chargingsettings_parser.add_argument("--target-soc", help="Desired charging target SoC", nargs="?", type=int)
    chargingsettings_parser.add_argument("--ac-limit", help="Maximum AC limit", nargs="?", type=int)
    chargingsettings_parser.set_defaults(func=chargingsettings)

    chargingprofile_parser = subparsers.add_parser("chargingprofile", description="Set vehicle charging profile.")
    _add_default_arguments(chargingprofile_parser)
    chargingprofile_parser.add_argument("vin", help=TEXT_VIN)
    chargingprofile_parser.add_argument(
        "--charging-mode",
        help="Desired charging mode",
        nargs="?",
        type=ChargingMode,
        choices=[cm.value for cm in ChargingMode if cm != ChargingMode.UNKNOWN],
    )
    chargingprofile_parser.add_argument(
        "--precondition-climate", help="Precondition climate on charging windows", nargs="?", type=bool
    )
    chargingprofile_parser.set_defaults(func=chargingprofile)

    charge_parser = subparsers.add_parser("charge", description="Start/stop charging on enabled vehicles.")
    _add_default_arguments(charge_parser)
    charge_parser.add_argument("vin", help=TEXT_VIN)
    charge_parser.add_argument("action", type=str, choices=["start", "stop"])
    charge_parser.set_defaults(func=charge)

    image_parser = subparsers.add_parser("image", description="Download a vehicle image.")
    _add_default_arguments(image_parser)
    image_parser.add_argument("vin", help=TEXT_VIN)
    image_parser.set_defaults(func=image)

    sendpoi_parser = subparsers.add_parser("sendpoi", description="Send a point of interest to the vehicle.")
    _add_default_arguments(sendpoi_parser)
    sendpoi_parser.add_argument("vin", help=TEXT_VIN)
    sendpoi_parser.add_argument("latitude", help="Latitude of the POI", type=float)
    sendpoi_parser.add_argument("longitude", help="Longitude of the POI", type=float)
    sendpoi_parser.add_argument("--name", help="Name of the POI", nargs="?", default=DEFAULT_POI_NAME)
    sendpoi_parser.add_argument(
        "--street", help="(optional, display only) Street & House No. of the POI", nargs="?", default=None
    )
    sendpoi_parser.add_argument("--city", help="(optional, display only) City of the POI", nargs="?", default=None)
    sendpoi_parser.add_argument(
        "--postalcode", help="(optional, display only) Postal code of the POI", nargs="?", default=None
    )
    sendpoi_parser.add_argument(
        "--country", help="(optional, display only) Country of the POI", nargs="?", default=None
    )
    sendpoi_parser.set_defaults(func=send_poi)

    sendpoi_from_address_parser = subparsers.add_parser(
        "sendpoi_from_address", description=("Send a point of interest parsed from a street address to the vehicle.")
    )
    _add_default_arguments(sendpoi_from_address_parser)
    sendpoi_from_address_parser.add_argument("vin", help=TEXT_VIN)
    sendpoi_from_address_parser.add_argument(
        "-n", "--name", help="(optional, display only) Name of the POI", nargs="?", default=None
    )
    sendpoi_from_address_parser.add_argument(
        "-a", "--address", nargs="+", help="Address (e.g. 'Street 17, city, zip, country')"
    )
    sendpoi_from_address_parser.set_defaults(func=send_poi_from_address)

    return parser


async def get_status(args) -> None:
    """Get the vehicle status."""
    if args.json:
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)

    account = MyBMWAccount(args.username, args.password, get_region_from_name(args.region))
    if args.lat and args.lng:
        account.set_observer_position(args.lat, args.lng)
    await account.get_vehicles()

    if args.json:
        print(json.dumps(account.vehicles, cls=MyBMWJSONEncoder))
    else:
        print(f"Found {len(account.vehicles)} vehicles: {','.join([v.name for v in account.vehicles])}")

        for vehicle in account.vehicles:
            print(f"VIN: {vehicle.vin}")
            print(f"Mileage: {vehicle.mileage.value} {vehicle.mileage.unit}")
            print("Vehicle data:")
            print(json.dumps(account.vehicles, cls=MyBMWJSONEncoder, indent=4))


def get_vehicle_or_return(account: MyBMWAccount, vin: str) -> MyBMWVehicle:
    """Get a vehicle by VIN or raise if not in account's vehicle list."""
    vehicle = account.get_vehicle(vin)
    if not vehicle:
        valid_vins = ", ".join(v.vin for v in account.vehicles)
        raise KeyError(f"Error: Could not find vehicle for VIN {vin}. Valid VINs are: {valid_vins}")
    return vehicle


async def fingerprint(args) -> None:
    """Save the vehicle fingerprint."""
    time_dir = Path.home() / "vehicle_fingerprint" / time.strftime("%Y-%m-%d_%H-%M-%S")
    time_dir.mkdir(parents=True)

    account = MyBMWAccount(
        args.username,
        args.password,
        get_region_from_name(args.region),
        log_responses=True,
    )
    if args.lat and args.lng:
        account.set_observer_position(args.lat, args.lng)
    await account.get_vehicles()

    log_response_store_to_file(account.get_stored_responses(), time_dir)
    print(f"fingerprint of the vehicles written to {time_dir}")


async def light_flash(args) -> None:
    """Trigger the vehicle to flash its lights."""
    account = MyBMWAccount(args.username, args.password, get_region_from_name(args.region))
    await account.get_vehicles()
    vehicle = get_vehicle_or_return(account, args.vin)
    status = await vehicle.remote_services.trigger_remote_light_flash()
    print(status.state)


async def horn(args) -> None:
    """Trigger the vehicle to horn."""
    account = MyBMWAccount(args.username, args.password, get_region_from_name(args.region))
    await account.get_vehicles()
    vehicle = get_vehicle_or_return(account, args.vin)
    status = await vehicle.remote_services.trigger_remote_horn()
    print(status.state)


async def vehicle_finder(args) -> None:
    """Trigger the vehicle finder to locate it."""
    account = MyBMWAccount(args.username, args.password, get_region_from_name(args.region))
    account.set_observer_position(args.lat, args.lng)
    await account.get_vehicles()
    vehicle = get_vehicle_or_return(account, args.vin)
    status = await vehicle.remote_services.trigger_remote_vehicle_finder()
    print(status.state)
    print({"gps_position": vehicle.vehicle_location.location, "heading": vehicle.vehicle_location.heading})


async def chargingsettings(args) -> None:
    """Trigger a change to charging settings."""
    if not args.target_soc and not args.ac_limit:
        raise ValueError("At least one of 'charging-target' and 'ac-limit' has to be provided.")
    account = MyBMWAccount(args.username, args.password, get_region_from_name(args.region))
    await account.get_vehicles()
    vehicle = get_vehicle_or_return(account, args.vin)
    status = await vehicle.remote_services.trigger_charging_settings_update(
        target_soc=args.target_soc, ac_limit=args.ac_limit
    )
    print(status.state)


async def chargingprofile(args) -> None:
    """Trigger a change to charging profile."""
    if not args.charging_mode and not args.precondition_climate:
        raise ValueError("At least one of 'charging-mode' and 'precondition-climate' has to be provided.")
    account = MyBMWAccount(args.username, args.password, get_region_from_name(args.region))
    await account.get_vehicles()
    vehicle = get_vehicle_or_return(account, args.vin)
    status = await vehicle.remote_services.trigger_charging_profile_update(
        charging_mode=args.charging_mode, precondition_climate=args.precondition_climate
    )
    print(status.state)


async def charge(args) -> None:
    """Trigger a vehicle to start or stop charging."""
    account = MyBMWAccount(args.username, args.password, get_region_from_name(args.region))
    await account.get_vehicles()
    vehicle = get_vehicle_or_return(account, args.vin)
    status = await getattr(vehicle.remote_services, f"trigger_charge_{args.action.lower()}")()
    print(status.state)


async def image(args) -> None:
    """Download a rendered image of the vehicle."""
    account = MyBMWAccount(args.username, args.password, get_region_from_name(args.region))
    await account.get_vehicles()
    vehicle = get_vehicle_or_return(account, args.vin)

    for viewdirection in VehicleViewDirection:
        if viewdirection == VehicleViewDirection.UNKNOWN:
            continue
        filename = str(viewdirection.name).lower() + ".png"
        with open(filename, "wb") as output_file:
            image_data = await vehicle.get_vehicle_image(viewdirection)
            output_file.write(image_data)
        print(f"vehicle image saved to {filename}")


async def send_poi(args) -> None:
    """Send Point Of Interest to car."""
    account = MyBMWAccount(args.username, args.password, get_region_from_name(args.region))
    await account.get_vehicles()
    vehicle = get_vehicle_or_return(account, args.vin)

    poi_data = {
        "lat": args.latitude,
        "lon": args.longitude,
        "name": args.name,
        "street": args.street,
        "city": args.city,
        "postal_code": args.postalcode,
        "country": args.country,
    }
    await vehicle.remote_services.trigger_send_poi(poi_data)


async def send_poi_from_address(args) -> None:
    """Create Point of Interest from OSM Nominatim and send to car."""
    account = MyBMWAccount(args.username, args.password, get_region_from_name(args.region))
    await account.get_vehicles()
    vehicle = get_vehicle_or_return(account, args.vin)

    query = [(str(" ".join(args.address)))]
    try:
        async with httpx.AsyncClient() as client:
            response: httpx.Response = await client.get(
                "https://nominatim.openstreetmap.org",
                params={"q": query, "format": "json", "addressdetails": 1, "limit": 1},
            )
            response_json = response.json()[0]
    except IndexError:
        print("\nAddress not found")
        sys.exit(1)
    address = response_json.get("address", {})
    city = address.get("city")
    town = address.get("town")

    poi_data = {
        "lat": response_json["lat"],
        "lon": response_json["lon"],
        "name": args.name,
        "street": address.get("road"),
        "city": town if city is None and town is not None else None,
        "postal_code": address.get("postcode"),
        "country": address.get("country"),
    }
    await vehicle.remote_services.trigger_send_poi(poi_data)


def _add_default_arguments(parser: argparse.ArgumentParser):
    """Add the default arguments username, password, region to the parser."""
    parser.add_argument("username", help="Connected Drive username")
    parser.add_argument("password", help="Connected Drive password")
    parser.add_argument("region", choices=valid_regions(), help="Region of the Connected Drive account")


def _add_position_arguments(parser: argparse.ArgumentParser):
    """Add the lat and lng attributes to the parser."""
    parser.add_argument("lat", type=float, nargs="?", const=0.0, help="(optional) Your current GPS latitude (as float)")
    parser.add_argument(
        "lng", type=float, nargs="?", const=0.0, help="(optional) Your current GPS longitude (as float)"
    )
    parser.set_defaults(func=get_status)


def main():
    """Get arguments from parser and run function in event loop."""
    parser = main_parser()
    args = parser.parse_args()

    loop = asyncio.get_event_loop()
    loop.run_until_complete(args.func(args))


if __name__ == "__main__":
    main()



"""General utils and base classes used in the library."""


import datetime
import inspect
import json
import logging
import pathlib
import time
from enum import Enum
from typing import TYPE_CHECKING, Dict, List, Optional, Union

from bimmer_connected.models import AnonymizedResponse

if TYPE_CHECKING:
    from typing import TypeVar

    from typing_extensions import ParamSpec

    _T = TypeVar("_T")
    _R = TypeVar("_R")
    _P = ParamSpec("_P")
_LOGGER = logging.getLogger(__name__)


JSON_IGNORED_KEYS = ["account", "_account", "vehicle", "_vehicle", "status", "remote_services"]


def get_class_property_names(obj: object):
    """Return the names of all properties of a class."""
    return [p[0] for p in inspect.getmembers(type(obj), inspect.isdatadescriptor) if not p[0].startswith("_")]


def parse_datetime(date_str: str) -> Optional[datetime.datetime]:
    """Convert a time string into datetime."""
    if not date_str:
        return None
    date_formats = ["%Y-%m-%dT%H:%M:%S.%f%z", "%Y-%m-%dT%H:%M:%S%z", "%Y-%m-%dT%H:%M:%S.%fZ", "%Y-%m-%dT%H:%M:%SZ"]
    for date_format in date_formats:
        try:
            # Parse datetimes using `time.strptime` to allow running in some embedded python interpreters.
            # https://bugs.python.org/issue27400
            time_struct = time.strptime(date_str, date_format)
            parsed = datetime.datetime(*(time_struct[0:6]))
            if time_struct.tm_gmtoff and time_struct.tm_gmtoff != 0:
                parsed = parsed - datetime.timedelta(seconds=time_struct.tm_gmtoff)
            parsed = parsed.replace(tzinfo=datetime.timezone.utc)
            return parsed
        except ValueError:
            pass
    _LOGGER.error("unable to parse '%s' using %s", date_str, date_formats)
    return None


class MyBMWJSONEncoder(json.JSONEncoder):
    """JSON Encoder that handles data classes, properties and additional data types."""

    def default(self, o) -> Union[str, dict]:  # noqa: D102
        if isinstance(o, (datetime.datetime, datetime.date, datetime.time)):
            return o.isoformat()
        if not isinstance(o, Enum) and hasattr(o, "__dict__") and isinstance(o.__dict__, Dict):
            retval: Dict = o.__dict__
            retval.update({p: getattr(o, p) for p in get_class_property_names(o)})
            return {k: v for k, v in retval.items() if k not in JSON_IGNORED_KEYS}
        return str(o)


def to_camel_case(input_str: str) -> str:
    """Convert SNAKE_CASE or snake_case to camelCase."""

    retval = ""
    flag_upper = False
    for curr in input_str.lower():
        if not curr.isalnum():
            if curr == "_":
                flag_upper = True
            continue
        retval = retval + (curr.upper() if flag_upper else curr)
        flag_upper = False
    return retval


def log_response_store_to_file(response_store: List[AnonymizedResponse], logfile_path: pathlib.Path) -> None:
    """Log all responses to files."""

    for response in response_store:
        output_path = logfile_path / response.filename
        content = response.content

        with open(output_path, "w", encoding="UTF-8") as logfile:
            if output_path.suffix == ".json" or not isinstance(content, str):
                json.dump(content or [], logfile, indent=4, sort_keys=True)
            else:
                logfile.write(content or "NO CONTENT")



"""Models the state of the vehicle tires."""

import datetime
from dataclasses import dataclass
from typing import Any, Dict, Optional

from bimmer_connected.const import ATTR_STATE
from bimmer_connected.models import StrEnum, VehicleDataBase


class ClimateActivityState(StrEnum):
    """Possible AC states."""

    COOLING = "COOLING"
    HEATING = "HEATING"
    INACTIVE = "INACTIVE"
    STANDBY = "STANDBY"
    UNKNOWN = "UNKNOWN"


@dataclass
class Climate(VehicleDataBase):
    """Provides an accessible version of `state.climateControlState`."""

    activity: ClimateActivityState = ClimateActivityState.UNKNOWN
    """Current climate activity state."""

    activity_end_time_no_tz: Optional[datetime.datetime] = None
    """Climatization end time w/o timezone."""

    account_timezone: datetime.timezone = datetime.timezone.utc

    @property
    def is_climate_on(self) -> bool:
        """Return True if climatization is active."""
        return self.activity in [ClimateActivityState.COOLING, ClimateActivityState.HEATING]

    @property
    def activity_end_time(self) -> Optional[datetime.datetime]:
        """Climatization end time."""
        if self.activity_end_time_no_tz:
            return self.activity_end_time_no_tz.astimezone(self.account_timezone)
        return None

    @classmethod
    def _parse_vehicle_data(cls, vehicle_data: Dict) -> Dict:
        """Parse tire status."""
        retval: Dict[str, Any] = {}

        if ATTR_STATE in vehicle_data:
            if "climateControlState" in vehicle_data[ATTR_STATE]:
                retval["activity"] = ClimateActivityState(vehicle_data[ATTR_STATE]["climateControlState"]["activity"])
                retval["activity_end_time_no_tz"] = (
                    (
                        datetime.datetime.now()
                        + datetime.timedelta(
                            seconds=int(vehicle_data[ATTR_STATE]["climateControlState"]["remainingSeconds"])
                        )
                    )
                    if "remainingSeconds" in vehicle_data[ATTR_STATE]["climateControlState"]
                    else None
                )

        return retval



"""Tests for utils."""
import datetime
import json

try:
    import zoneinfo
except ImportError:
    from backports import zoneinfo  # type: ignore[import, no-redef]

import pytest
import respx
import time_machine

from bimmer_connected.api.utils import get_capture_position
from bimmer_connected.models import ChargingSettings, ValueWithUnit
from bimmer_connected.utils import MyBMWJSONEncoder, get_class_property_names, parse_datetime

from . import RESPONSE_DIR, VIN_G26, load_response
from .conftest import prepare_account_with_vehicles


@pytest.mark.asyncio
async def test_drive_train(bmw_fixture: respx.Router):
    """Tests available attribute."""
    vehicle = (await prepare_account_with_vehicles()).get_vehicle(VIN_G26)
    assert [
        "available_attributes",
        "brand",
        "drive_train",
        "drive_train_attributes",
        "has_combustion_drivetrain",
        "has_electric_drivetrain",
        "is_charging_plan_supported",
        "is_lsc_enabled",
        "is_remote_charge_start_enabled",
        "is_remote_charge_stop_enabled",
        "is_remote_climate_start_enabled",
        "is_remote_climate_stop_enabled",
        "is_remote_horn_enabled",
        "is_remote_lights_enabled",
        "is_remote_lock_enabled",
        "is_remote_sendpoi_enabled",
        "is_remote_set_ac_limit_enabled",
        "is_remote_set_target_soc_enabled",
        "is_remote_unlock_enabled",
        "is_vehicle_active",
        "is_vehicle_tracking_enabled",
        "lsc_type",
        "mileage",
        "name",
        "timestamp",
        "vin",
    ] == get_class_property_names(vehicle)


def test_parse_datetime(caplog):
    """Test datetime parser."""

    dt_without_milliseconds = datetime.datetime(2021, 11, 12, 13, 14, 15, tzinfo=datetime.timezone.utc)

    assert dt_without_milliseconds == parse_datetime("2021-11-12T13:14:15.567Z")

    assert dt_without_milliseconds == parse_datetime("2021-11-12T13:14:15Z")

    assert dt_without_milliseconds == parse_datetime("2021-11-12T16:14:15+03:00")

    unparseable_datetime = "2021-14-12T13:14:15Z"
    assert parse_datetime(unparseable_datetime) is None
    errors = [r for r in caplog.records if r.levelname == "ERROR" and unparseable_datetime in r.message]
    assert len(errors) == 1


@time_machine.travel(
    datetime.datetime(2011, 11, 28, tzinfo=zoneinfo.ZoneInfo("America/Los_Angeles")),
    tick=False,
)
@pytest.mark.asyncio
async def test_account_timezone(bmw_fixture: respx.Router):
    """Test the timezone in MyBMWAccount."""
    account = await prepare_account_with_vehicles()
    assert account.utcdiff == 960


def test_json_encoder():
    """Test the MyBMWJSONEncoder."""
    encoded = json.dumps(
        {
            "datetime": datetime.datetime(2022, 6, 2, 22, 19, 34, 123456),
            "date": datetime.date(2022, 6, 2),
            "value": ValueWithUnit(17, "mi"),
            "list": [
                {
                    "value_int": 1,
                    "value_str": "string",
                },
                zoneinfo.ZoneInfo("America/Los_Angeles"),
            ],
        },
        cls=MyBMWJSONEncoder,
    )

    assert (
        '{"datetime": "2022-06-02T22:19:34.123456", "date": "2022-06-02", "value": [17, "mi"],'
        ' "list": [{"value_int": 1, "value_str": "string"}, "America/Los_Angeles"]}'
    ) == encoded


def test_charging_settings():
    """Test parsing and validation of charging settings."""

    cs = ChargingSettings(chargingTarget=90, acLimitValue=32)
    assert cs.acLimitValue == 32
    assert cs.chargingTarget == 90
    assert cs.dcLoudness is None
    assert cs.isUnlockCableActive is None


def test_get_capture_position():
    """Test the auto get slider captcha position."""
    base64_background_img = load_response(RESPONSE_DIR / "auth" / "auth_slider_captcha.json")["data"]["backGroundImg"]
    position = get_capture_position(base64_background_img)
    assert position == "0.81"



"""URLs for different services and error code mapping."""
from enum import Enum


class CarBrands(str, Enum):
    """Car brands supported by the MyBMW API."""

    @classmethod
    def _missing_(cls, value):
        value = next(iter(value.split("_")))
        for member in cls:
            if member.value == value.lower():
                return member
        raise ValueError(f"'{value}' is not a valid {cls.__name__}")

    BMW = "bmw"
    MINI = "mini"


class Regions(str, Enum):
    """Regions of the world with separate servers."""

    NORTH_AMERICA = "na"
    CHINA = "cn"
    REST_OF_WORLD = "row"


SERVER_URLS_MYBMW = {
    Regions.NORTH_AMERICA: "cocoapi.bmwgroup.us",
    Regions.REST_OF_WORLD: "cocoapi.bmwgroup.com",
    Regions.CHINA: "myprofile.bmw.com.cn",
}

OCP_APIM_KEYS = {
    Regions.NORTH_AMERICA: "MzFlMTAyZjUtNmY3ZS03ZWYzLTkwNDQtZGRjZTYzODkxMzYy",
    Regions.REST_OF_WORLD: "NGYxYzg1YTMtNzU4Zi1hMzdkLWJiYjYtZjg3MDQ0OTRhY2Zh",
}

APP_VERSIONS = {
    Regions.NORTH_AMERICA: "3.9.0(27760)",
    Regions.REST_OF_WORLD: "3.9.0(27760)",
    Regions.CHINA: "3.6.1(23634)",
}

HTTPX_TIMEOUT = 30.0

USER_AGENTS = {
    Regions.NORTH_AMERICA: "Dart/2.19 (dart:io)",
    Regions.REST_OF_WORLD: "Dart/2.19 (dart:io)",
    Regions.CHINA: "Dart/2.18 (dart:io)",
}
X_USER_AGENT = "android(TQ2A.230405.003.B2);{brand};{app_version};{region}"


AUTH_CHINA_PUBLIC_KEY_URL = "/eadrax-coas/v1/cop/publickey"
AUTH_CHINA_LOGIN_URL = "/eadrax-coas/v2/login/pwd"
AUTH_CHINA_TOKEN_URL = "/eadrax-coas/v2/oauth/token"
AUTH_CHINA_CAPTCHA_URL = "/eadrax-coas/v2/cop/slider-captcha"
AUTH_CHINA_CAPTCHA_CHECK_URL = "/eadrax-coas/v1/cop/check-captcha"

OAUTH_CONFIG_URL = "/eadrax-ucs/v1/presentation/oauth/config"

VEHICLES_URL = "/eadrax-vcs/v4/vehicles"
VEHICLE_STATE_URL = VEHICLES_URL + "/state"

REMOTE_SERVICE_BASE_URL = "/eadrax-vrccs/v3/presentation/remote-commands"
REMOTE_SERVICE_URL = REMOTE_SERVICE_BASE_URL + "/{vin}/{service_type}"
REMOTE_SERVICE_STATUS_URL = REMOTE_SERVICE_BASE_URL + "/eventStatus?eventId={event_id}"
REMOTE_SERVICE_POSITION_URL = REMOTE_SERVICE_BASE_URL + "/eventPosition?eventId={event_id}"

VEHICLE_CHARGING_DETAILS_URL = "/eadrax-crccs/v2/vehicles"
VEHICLE_CHARGING_BASE_URL = "/eadrax-crccs/v1/vehicles/{vin}"
VEHICLE_CHARGING_SETTINGS_SET_URL = VEHICLE_CHARGING_BASE_URL + "/charging-settings"
VEHICLE_CHARGING_PROFILE_SET_URL = VEHICLE_CHARGING_BASE_URL + "/charging-profile"
VEHICLE_CHARGING_START_STOP_URL = VEHICLE_CHARGING_BASE_URL + "/{service_type}"

VEHICLE_IMAGE_URL = "/eadrax-ics/v3/presentation/vehicles/{vin}/images?carView={view}"
VEHICLE_POI_URL = "/eadrax-dcs/v1/send-to-car/send-to-car"

VEHICLE_CHARGING_STATISTICS_URL = "/eadrax-chs/v1/charging-statistics"
VEHICLE_CHARGING_SESSIONS_URL = "/eadrax-chs/v1/charging-sessions"

SERVICE_CHARGING_STATISTICS_URL = "CHARGING_STATISTICS"
SERVICE_CHARGING_SESSIONS_URL = "CHARGING_SESSIONS"
SERVICE_CHARGING_PROFILE = "CHARGING_PROFILE"


ATTR_STATE = "state"
ATTR_CAPABILITIES = "capabilities"
ATTR_ATTRIBUTES = "attributes"
ATTR_CHARGING_SETTINGS = "charging_settings"

DEFAULT_POI_NAME = "Sent with ♥ by bimmer_connected"



"""Vehicle-wide constants used by multiple data models."""

from bimmer_connected.models import StrEnum


class DriveTrainType(StrEnum):
    """Different types of drive trains."""

    COMBUSTION = "COMBUSTION"
    PLUGIN_HYBRID = "PLUGIN_HYBRID"  # PHEV
    ELECTRIC = "ELECTRIC"
    ELECTRIC_WITH_RANGE_EXTENDER = "ELECTRIC_WITH_RANGE_EXTENDER"
    HYBRID = "HYBRID"  # mild hybrids (MyBMW API v1)
    MILD_HYBRID = "MILD_HYBRID"  # mild hybrids (MyBMW API v2)
    UNKNOWN = "UNKNOWN"


#: Set of drive trains that have a combustion engine
COMBUSTION_ENGINE_DRIVE_TRAINS = {
    DriveTrainType.COMBUSTION,
    DriveTrainType.ELECTRIC_WITH_RANGE_EXTENDER,
    DriveTrainType.PLUGIN_HYBRID,
    DriveTrainType.HYBRID,
    DriveTrainType.MILD_HYBRID,
}

#: set of drive trains that have a high voltage battery
HV_BATTERY_DRIVE_TRAINS = {
    DriveTrainType.PLUGIN_HYBRID,
    DriveTrainType.ELECTRIC,
    DriveTrainType.ELECTRIC_WITH_RANGE_EXTENDER,
}
